"""Tests for file operation functionality in tools.py"""

import os
import shutil
import tempfile
import unittest
from tempfile import gettempdir
from unittest.mock import MagicMock, mock_open, patch

from wcgw.client.tools import (
    BASH_STATE,
    get_context_for_errors,
    read_image_from_shell,
    save_out_of_context,
    truncate_if_over,
    write_file,
)
from wcgw.types_ import WriteIfEmpty


class TestFileOperations(unittest.TestCase):
    def setUp(self):
        """Set up test environment before each test case"""
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp()
        self.test_file_path = os.path.join(self.test_dir, "test_file.py")
        BASH_STATE.add_to_whitelist_for_overwrite(self.test_file_path)
        # Mock command_run for Docker tests
        self.mock_command_run_patcher = patch("wcgw.client.tools.command_run")
        self.mock_command_run = self.mock_command_run_patcher.start()
        self.mock_command_run.return_value = (0, "", "")

    def tearDown(self):
        """Clean up test environment after each test case"""
        self.mock_command_run_patcher.stop()
        shutil.rmtree(self.test_dir)
        BASH_STATE.set_in_docker("")

    def test_truncate_if_over(self):
        """Test content truncation based on token limit"""
        content = "test" * 100

        # Test with no limit
        result = truncate_if_over(content, None)
        self.assertEqual(result, content)

        # Test with limit
        with patch("wcgw.client.tools.default_enc") as mock_enc:
            # Create mock token list that exceeds max_tokens limit
            token_ids = list(range(1000))  # 1000 tokens to ensure truncation
            mock_token_list = MagicMock()
            mock_token_list.ids = token_ids
            mock_token_list.__len__ = MagicMock(return_value=len(token_ids))
            mock_enc.encode.return_value = mock_token_list

            # Set up the decode behavior to return "truncated beginning" when truncated
            def mock_decode(token_list):
                if len(token_list) < len(token_ids):
                    return "truncated beginning"
                return content

            mock_enc.decode.side_effect = mock_decode

            # Test truncation with max_tokens=150 to trigger truncation
            result = truncate_if_over(content, 150)

            # Should truncate tokens to max(0, 150-100) = 50 and append truncation marker
            expected = "truncated beginning\n(...truncated)"
            self.assertEqual(result, expected)

            # Verify encode was called with original content
            mock_enc.encode.assert_called_once_with(content)

            # Verify decode was called with truncated token list
            mock_enc.decode.assert_called_once_with(
                token_ids[:50]
            )  # max(0, 150-100) = 50

    def test_read_image_from_shell(self):
        """Test read_image_from_shell functionality"""
        # Test regular file read
        with patch("os.path.exists", return_value=True):
            with patch("os.path.isabs", return_value=True):
                with patch("builtins.open", mock_open(read_data=b"test_image_data")):
                    result = read_image_from_shell("/test/image.png")
                    self.assertEqual(result.media_type, "image/png")
                    self.assertEqual(
                        result.data, "dGVzdF9pbWFnZV9kYXRh"
                    )  # base64 of "test_image_data"

        # Test non-existent file
        with patch("os.path.exists", return_value=False):
            with patch("os.path.isabs", return_value=True):
                with self.assertRaises(ValueError):
                    read_image_from_shell("/nonexistent/image.png")

    def test_get_context_for_errors(self):
        """Test get_context_for_errors functionality"""
        file_content = "line1\nline2\nline3\nline4\nline5"
        errors = [(2, 0)]  # Error on line 2

        context = get_context_for_errors(errors, file_content, None)
        self.assertIn("line2", context)
        self.assertIn("```", context)

        # Test with small token limit
        short_context = get_context_for_errors(errors, file_content, 10)
        self.assertEqual(
            short_context, "Please re-read the file to understand the context"
        )

    def test_save_out_of_context(self):
        """Test save_out_of_context functionality"""
        content = "Test content"
        suffix = ".txt"

        filepath = save_out_of_context(content, suffix)
        try:
            # Verify file exists and has correct content
            self.assertTrue(os.path.exists(filepath))
            self.assertTrue(filepath.startswith(gettempdir()))
            self.assertTrue(filepath.endswith(suffix))
            with open(filepath, "r") as f:
                self.assertEqual(f.read(), content)
        finally:
            # Cleanup
            if os.path.exists(filepath):
                os.remove(filepath)

    def test_write_file_syntax_check(self):
        """Test write_file syntax checking"""
        # Create mock syntax checker
        mock_checker = MagicMock()
        mock_checker.description = "Invalid syntax"
        mock_checker.errors = [(1, 0)]  # Error at line 1, col 0

        # Setup test file
        mock_encoder = MagicMock()
        mock_encoder.encode.return_value.ids = [1, 2, 3]  # Mock token encoding
        mock_encoder.decode.return_value = "mocked content"

        with patch("wcgw.client.tools.check_syntax", return_value=mock_checker):
            with patch("wcgw.client.tools.default_enc", mock_encoder):
                write_arg = WriteIfEmpty(
                    file_path=self.test_file_path, file_content="invalid code"
                )
                with patch("pathlib.Path.exists", return_value=False):
                    result = write_file(write_arg, error_on_exist=True, max_tokens=100)
                    self.assertIn("Success", result)
                    self.assertIn("syntax errors", result)
                    self.assertIn("Invalid syntax", result)
                    self.assertIn("tree-sitter reported syntax errors", result)


if __name__ == "__main__":
    unittest.main()



import pytest
from typing import cast
from openai.types.chat import ChatCompletionMessage, ChatCompletionMessageParam, ParsedChatCompletionMessage
from tokenizers import Tokenizer
from wcgw.client.openai_utils import get_input_cost, get_output_cost
from wcgw.client.common import CostData


@pytest.fixture
def tokenizer():
    # Create a simple tokenizer for testing
    return Tokenizer.from_pretrained("gpt2")


@pytest.fixture
def cost_data():
    return CostData(
        cost_per_1m_input_tokens=0.01,
        cost_per_1m_output_tokens=0.03
    )


def test_get_input_cost_simple(tokenizer, cost_data):
    history = [
        {"role": "user", "content": "Hello"},
        {"role": "assistant", "content": "Hi there"}
    ]
    cost, tokens = get_input_cost(cost_data, tokenizer, history)
    assert tokens > 0
    assert cost > 0


def test_get_input_cost_with_list_content(tokenizer, cost_data):
    history = [
        {"role": "user", "content": [{"type": "text", "text": "Hello"}]},
        {"role": "assistant", "content": "Hi there"}
    ]
    cost, tokens = get_input_cost(cost_data, tokenizer, history)
    assert tokens > 0
    assert cost > 0


def test_get_input_cost_with_refusal(tokenizer, cost_data):
    history = [
        {"role": "user", "content": None, "refusal": "Content refused"},
        {"role": "assistant", "content": "Hi there"}
    ]
    cost, tokens = get_input_cost(cost_data, tokenizer, history)
    assert tokens > 0
    assert cost > 0


def test_get_input_cost_invalid_content():
    cost_data = CostData(cost_per_1m_input_tokens=0.01, cost_per_1m_output_tokens=0.03)
    tokenizer = Tokenizer.from_pretrained("gpt2")
    history = [{"role": "user", "content": 123}]  # type: ignore
    
    with pytest.raises(ValueError, match="Expected content to be string"):
        get_input_cost(cost_data, tokenizer, history)


def test_get_output_cost_simple_message(tokenizer, cost_data):
    message = ChatCompletionMessage(
        content="Hello world",
        role="assistant",
        tool_calls=None
    )
    cost, tokens = get_output_cost(cost_data, tokenizer, message)
    assert tokens > 0
    assert cost > 0


def test_get_output_cost_with_tool_calls(tokenizer, cost_data):
    message = {
        "role": "assistant",
        "content": "Let me help you with that",
        "tool_calls": [
            {
                "id": "call_123",
                "type": "function",
                "function": {
                    "name": "test_function",
                    "arguments": '{"arg1": "test"}'
                }
            }
        ]
    }
    cost, tokens = get_output_cost(cost_data, tokenizer, cast(ChatCompletionMessageParam, message))
    assert tokens > 0
    assert cost > 0


def test_get_output_cost_tool_message(tokenizer, cost_data):
    message = {
        "role": "tool",
        "content": "Tool execution result",
        "tool_call_id": "call_123"
    }
    cost, tokens = get_output_cost(cost_data, tokenizer, cast(ChatCompletionMessageParam, message))
    assert cost == 0
    assert tokens == 0


import anyio
import click
import mcp_wcgw.types as types
from mcp_wcgw.server import Server


def create_messages(
    context: str | None = None, topic: str | None = None
) -> list[types.PromptMessage]:
    """Create the messages for the prompt."""
    messages = []

    # Add context if provided
    if context:
        messages.append(
            types.PromptMessage(
                role="user",
                content=types.TextContent(
                    type="text", text=f"Here is some relevant context: {context}"
                ),
            )
        )

    # Add the main prompt
    prompt = "Please help me with "
    if topic:
        prompt += f"the following topic: {topic}"
    else:
        prompt += "whatever questions I may have."

    messages.append(
        types.PromptMessage(
            role="user", content=types.TextContent(type="text", text=prompt)
        )
    )

    return messages


@click.command()
@click.option("--port", default=8000, help="Port to listen on for SSE")
@click.option(
    "--transport",
    type=click.Choice(["stdio", "sse"]),
    default="stdio",
    help="Transport type",
)
def main(port: int, transport: str) -> int:
    app = Server("mcp-simple-prompt")

    @app.list_prompts()
    async def list_prompts() -> list[types.Prompt]:
        return [
            types.Prompt(
                name="simple",
                description="A simple prompt that can take optional context and topic "
                "arguments",
                arguments=[
                    types.PromptArgument(
                        name="context",
                        description="Additional context to consider",
                        required=False,
                    ),
                    types.PromptArgument(
                        name="topic",
                        description="Specific topic to focus on",
                        required=False,
                    ),
                ],
            )
        ]

    @app.get_prompt()
    async def get_prompt(
        name: str, arguments: dict[str, str] | None = None
    ) -> types.GetPromptResult:
        if name != "simple":
            raise ValueError(f"Unknown prompt: {name}")

        if arguments is None:
            arguments = {}

        return types.GetPromptResult(
            messages=create_messages(
                context=arguments.get("context"), topic=arguments.get("topic")
            ),
            description="A simple prompt with optional context and topic arguments",
        )

    if transport == "sse":
        from mcp_wcgw.server.sse import SseServerTransport
        from starlette.applications import Starlette
        from starlette.routing import Route

        sse = SseServerTransport("/messages")

        async def handle_sse(request):
            async with sse.connect_sse(
                request.scope, request.receive, request._send
            ) as streams:
                await app.run(
                    streams[0], streams[1], app.create_initialization_options()
                )

        async def handle_messages(request):
            await sse.handle_post_message(request.scope, request.receive, request._send)

        starlette_app = Starlette(
            debug=True,
            routes=[
                Route("/sse", endpoint=handle_sse),
                Route("/messages", endpoint=handle_messages, methods=["POST"]),
            ],
        )

        import uvicorn

        uvicorn.run(starlette_app, host="0.0.0.0", port=port)
    else:
        from mcp_wcgw.server.stdio import stdio_server

        async def arun():
            async with stdio_server() as streams:
                await app.run(
                    streams[0], streams[1], app.create_initialization_options()
                )

        anyio.run(arun)

    return 0



import unittest
from unittest.mock import patch, mock_open, MagicMock
import os
import tempfile
from pathlib import Path
from wcgw.client.tools import (
    write_file,
    read_image_from_shell,
    get_context_for_errors,
    save_out_of_context,
    truncate_if_over,
    ImageData,
    BASH_STATE,
)
from wcgw.types_ import WriteIfEmpty

class TestToolsFiles(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        BASH_STATE._is_in_docker = ""  # Reset docker state
        BASH_STATE._whitelist_for_overwrite.clear()  # Clear whitelist
        self.temp_dir = tempfile.mkdtemp()  # Create a temporary directory

    def tearDown(self):
        # Clean up temporary directory recursively 
        if hasattr(self, 'temp_dir') and os.path.exists(self.temp_dir):
            for root, dirs, files in os.walk(self.temp_dir, topdown=False):
                for name in files:
                    os.remove(os.path.join(root, name))
                for name in dirs:
                    os.rmdir(os.path.join(root, name))
            os.rmdir(self.temp_dir)


    def test_write_file_validation(self):
        """Test write_file input validation"""
        # Test relative path
        test_file = WriteIfEmpty(file_path="relative/path.txt", file_content="test")
        result = write_file(test_file, error_on_exist=True, max_tokens=100)
        self.assertIn("Failure: file_path should be absolute path", result)

        # Test with empty content
        test_file = WriteIfEmpty(file_path=os.path.join(self.temp_dir, "test.txt"), file_content="")
        with patch('os.path.isabs', return_value=True):
            with patch('os.path.exists', return_value=False):
                result = write_file(test_file, error_on_exist=True, max_tokens=100)
                self.assertIn("Success", result)

    def test_write_file_existing(self):
        """Test write_file behavior with existing files"""
        test_file = WriteIfEmpty(file_path=os.path.join(self.temp_dir, "file.txt"), file_content="new content")
        
        with patch('os.path.isabs', return_value=True):
            with patch('os.path.exists', return_value=True):
                with patch('pathlib.Path.read_text', return_value="existing content"):
                    # Test with error_on_exist=True
                    result = write_file(test_file, error_on_exist=True, max_tokens=100)
                    self.assertIn("Error: can't write to existing file", result)
                    self.assertIn("existing content", result)

                    # Test with whitelisted file
                    BASH_STATE.add_to_whitelist_for_overwrite(os.path.join(self.temp_dir, "file.txt"))
                    with patch('pathlib.Path.open', mock_open()):
                        result = write_file(test_file, error_on_exist=True, max_tokens=100)
                        self.assertIn("Success", result)

    @patch('wcgw.client.tools.check_syntax')
    def test_write_file_syntax_check(self, mock_check):
        """Test write_file syntax checking"""
        # Setup syntax error mock
        mock_error = MagicMock()
        mock_error.description = "Invalid syntax"
        mock_error.errors = [(1, 0)]
        mock_check.return_value = mock_error

        test_file = WriteIfEmpty(file_path=os.path.join(self.temp_dir, "file.py"), file_content="invalid python")
        
        with patch('os.path.isabs', return_value=True):
            with patch('os.path.exists', return_value=False):
                with patch('pathlib.Path.open', mock_open()):
                    result = write_file(test_file, error_on_exist=True, max_tokens=100)
                    self.assertIn("Success", result)
                    self.assertIn("syntax errors", result)
                    self.assertIn("Invalid syntax", result)

    @patch('os.path.exists')
    @patch('os.path.isabs')
    @patch('builtins.open', new_callable=mock_open)
    def test_read_image_from_shell(self, mock_file, mock_isabs, mock_exists):
        """Test read_image_from_shell functionality"""
        # Setup mocks
        mock_isabs.return_value = True
        mock_exists.return_value = True
        mock_file.return_value.read.return_value = b"test_image_data"
        
        # Test regular file read
        result = read_image_from_shell("/test/image.png")
        self.assertIsInstance(result, ImageData)
        self.assertEqual(result.media_type, "image/png")
        
        # Test non-existent file
        mock_exists.return_value = False
        with self.assertRaises(ValueError):
            read_image_from_shell("/nonexistent/image.png")

    @patch('wcgw.client.tools.default_enc')
    def test_get_context_for_errors(self, mock_enc):
        """Test get_context_for_errors functionality"""
        mock_enc.encode.return_value = [1, 2, 3]  # simulate tokens
        
        # Test with single error
        file_content = "line1\nline2\nline3\nline4\nline5"
        errors = [(2, 0)]  # Error on line 2
        result = get_context_for_errors(errors, file_content, max_tokens=100)
        self.assertIn("line2", result)
        self.assertIn("```", result)

        # Test with multiple errors
        errors = [(2, 0), (4, 0)]
        result = get_context_for_errors(errors, file_content, max_tokens=100)
        


"""Tests for computer_use.py shell and screenshot functionality"""

import pytest
from unittest.mock import patch, MagicMock
from wcgw.client.computer_use import ComputerTool, ToolError


@pytest.fixture
def mock_command_run():
    with patch("wcgw.client.computer_use.command_run") as mock:
        mock.return_value = (0, "test output", "test error")
        yield mock


@pytest.fixture
def initialized_tool():
    tool = ComputerTool()
    tool.docker_image_id = "test_docker_id"
    tool.width = 1920
    tool.height = 1080
    tool._display_prefix = "DISPLAY=:0 "
    tool.xdotool = "DISPLAY=:0 xdotool"
    return tool


def test_get_screen_info_success(mock_command_run):
    mock_command_run.return_value = (0, "1920,1080,0\n", "")
    tool = ComputerTool()
    tool.docker_image_id = "test_docker_id"

    width, height, display_num = tool.get_screen_info()
    
    assert width == 1920
    assert height == 1080
    assert display_num == 0
    assert tool.width == 1920
    assert tool.height == 1080
    assert tool.display_num == 0
    assert tool._display_prefix == "DISPLAY=:0 "
    assert tool.xdotool == "DISPLAY=:0 xdotool"


def test_get_screen_info_no_display(mock_command_run):
    mock_command_run.return_value = (0, "1920,1080,\n", "")
    tool = ComputerTool()
    tool.docker_image_id = "test_docker_id"

    width, height, display_num = tool.get_screen_info()
    
    assert width == 1920
    assert height == 1080
    assert display_num is None
    assert tool._display_prefix == ""


def test_get_screen_info_defaults(mock_command_run):
    mock_command_run.return_value = (0, ",,\n", "")
    tool = ComputerTool()
    tool.docker_image_id = "test_docker_id"

    width, height, display_num = tool.get_screen_info()
    
    assert width == 1080  # Default width
    assert height == 1920  # Default height
    assert display_num is None


def test_shell_command(initialized_tool, mock_command_run):
    result = initialized_tool.shell("test command", take_screenshot=False)
    
    assert any(
        'docker exec test_docker_id bash -c \'test command\'' in str(call)
        for call in mock_command_run.call_args_list
    )
    
    assert result.output == "test output"
    assert result.error == "test error"
    assert result.base64_image is None


@patch("time.sleep")
def test_shell_command_with_screenshot(mock_sleep, initialized_tool, mock_command_run):
    with patch.object(initialized_tool, 'screenshot') as mock_screenshot:
        mock_screenshot.return_value.base64_image = "test_base64"
        
        result = initialized_tool.shell("test command", take_screenshot=True)
        
        assert mock_command_run.call_count >= 1
        assert mock_screenshot.call_count >= 1
        mock_sleep.assert_called_once_with(initialized_tool._screenshot_delay)
        
        assert result.output == "test output"
        assert result.error == "test error"
        assert result.base64_image == "test_base64"


@patch("os.path.exists")
@patch("builtins.open")
def test_screenshot_success(mock_open, mock_exists, initialized_tool, mock_command_run):
    mock_exists.return_value = True
    mock_file = MagicMock()
    mock_file.__enter__.return_value.read.return_value = b"test image data"
    mock_open.return_value = mock_file

    result = initialized_tool.screenshot()

    # Check mkdir command
    assert any("mkdir -p /tmp/outputs" in str(call) for call in mock_command_run.call_args_list)
    # Check scrot command
    assert any("scrot -f" in str(call) for call in mock_command_run.call_args_list)
    
    assert result.output == ""
    assert result.base64_image == "dGVzdCBpbWFnZSBkYXRh"  # base64 encoded "test image data"


def test_screenshot_failure(initialized_tool, mock_command_run):
    mock_command_run.return_value = (1, "", "Screenshot failed")

    with pytest.raises(ToolError, match="Failed to take screenshot: Screenshot failed"):
        initialized_tool.screenshot()



from dataclasses import dataclass
from typing import Generic, TypeVar

from mcp_wcgw.shared.session import BaseSession
from mcp_wcgw.types import RequestId, RequestParams

SessionT = TypeVar("SessionT", bound=BaseSession)


@dataclass
class RequestContext(Generic[SessionT]):
    request_id: RequestId
    meta: RequestParams.Meta | None
    session: SessionT



import pytest
from wcgw.client.repo_ops.path_prob import FastPathAnalyzer
from unittest.mock import patch, mock_open, MagicMock


@pytest.fixture
def mock_tokenizer():
    tokenizer = MagicMock()
    encoding = MagicMock()
    encoding.tokens = ["test", "path", "token"]
    tokenizer.encode_batch.return_value = [encoding]
    tokenizer.decode.return_value = "test/path/token"
    return tokenizer


@pytest.fixture
def sample_vocab_content():
    return """test -1.5
path -2.0
token -1.8
unknown -3.0"""


def test_initialization(sample_vocab_content):
    with patch('builtins.open', mock_open(read_data=sample_vocab_content)):
        with patch('tokenizers.Tokenizer.from_file') as mock_tokenizer:
            analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
            
            # Check if vocab was loaded correctly
            assert len(analyzer.vocab_probs) == 4
            assert analyzer.vocab_probs["test"] == -1.5
            assert analyzer.vocab_probs["path"] == -2.0
            assert analyzer.vocab_probs["token"] == -1.8
            assert analyzer.vocab_probs["unknown"] == -3.0


def test_tokenize_batch(mock_tokenizer):
    with patch('builtins.open', mock_open()):
        with patch('tokenizers.Tokenizer.from_file', return_value=mock_tokenizer):
            analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
            
            # When we call tokenize_batch
            result = analyzer.tokenize_batch(["test/path"])
            assert result == [["test", "path", "token"]]
            mock_tokenizer.encode_batch.assert_called_once_with(["test/path"])


def test_detokenize(mock_tokenizer):
    with patch('builtins.open', mock_open()):
        with patch('tokenizers.Tokenizer.from_file', return_value=mock_tokenizer):
            analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
            
            result = analyzer.detokenize(["test", "path", "token"])
            assert result == "test/path/token"
            mock_tokenizer.decode.assert_called_once()


def test_calculate_path_probability(sample_vocab_content, mock_tokenizer):
    mock_tokenizer.encode_batch.return_value = [MagicMock(tokens=["test", "path", "token"])]
    
    with patch('builtins.open', mock_open(read_data=sample_vocab_content)):
        with patch('tokenizers.Tokenizer.from_file', return_value=mock_tokenizer):
            analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
            
            log_prob, tokens, unknown = analyzer.calculate_path_probability("test/path/token")
            
            # Expected probability: -1.5 + -2.0 + -1.8 = -5.3
            assert log_prob == -5.3
            assert tokens == ["test", "path", "token"]
            assert unknown == []


def test_calculate_path_probability_with_unknown_tokens(sample_vocab_content, mock_tokenizer):
    # Mock tokenizer to return some unknown tokens
    mock_tokenizer.encode_batch.return_value = [MagicMock(tokens=["test", "unknown_token", "path"])]
    
    with patch('builtins.open', mock_open(read_data=sample_vocab_content)):
        with patch('tokenizers.Tokenizer.from_file', return_value=mock_tokenizer):
            analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
            
            log_prob, tokens, unknown = analyzer.calculate_path_probability("test/unknown/path")
            
            # Expected probability: -1.5 + -2.0 = -3.5 (unknown token probability not included)
            assert log_prob == -3.5
            assert tokens == ["test", "unknown_token", "path"]
            assert unknown == ["unknown_token"]


def test_invalid_vocab_line():
    invalid_vocab = """test -1.5
invalid_line
path -2.0"""
    
    with patch('builtins.open', mock_open(read_data=invalid_vocab)):
        with patch('tokenizers.Tokenizer.from_file') as mock_tokenizer:
            analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
            
            # Check if vocab was loaded correctly, skipping invalid line
            assert len(analyzer.vocab_probs) == 2
            assert analyzer.vocab_probs["test"] == -1.5
            assert analyzer.vocab_probs["path"] == -2.0


    def test_empty_path():
        with patch('builtins.open', mock_open(read_data="")):
            with patch('tokenizers.Tokenizer.from_file') as mock_tokenizer:
                mock_tokenizer_instance = Mock()
                mock_tokenizer_instance.encode.return_value.tokens = []
                mock_tokenizer.return_value = mock_tokenizer_instance
                
                analyzer = FastPathAnalyzer("model.bin", "vocab.txt")
                log_prob, tokens, unknown = analyzer.calculate_path_probability("")
                assert log_prob == 0.0
                assert tokens == []
            assert unknown == []


import json
import os
import pytest
from pathlib import Path
import tempfile
from wcgw.client.anthropic_client import (
    text_from_editor,
    save_history,
    parse_user_message_special,
)
import rich.console
from unittest.mock import patch, mock_open, MagicMock


@pytest.fixture
def console():
    return rich.console.Console(style="bright_black", highlight=False, markup=False)


def test_text_from_editor_direct_input(console):
    with patch('builtins.input', return_value="Test input"):
        result = text_from_editor(console)
        assert result == "Test input"


def test_text_from_editor_with_editor():
    with tempfile.NamedTemporaryFile(mode='w', suffix='.tmp') as tf:
        tf.write("Test content from editor")
        tf.flush()
        
        with patch('builtins.input', return_value=""):
            with patch('os.environ.get', return_value="cat"):  # Use 'cat' as editor for testing
                with patch('subprocess.run') as mock_run:
                    with patch('builtins.open', mock_open(read_data="Test content from editor")):
                        console = rich.console.Console()
                        result = text_from_editor(console)
                        assert result == "Test content from editor"
                        mock_run.assert_called_once()


def test_save_history(tmp_path):
    # Create a temporary directory for testing
    history = [
        {"role": "system", "content": "System message"},
        {"role": "user", "content": "Test message"},
        {"role": "assistant", "content": "Test response"}
    ]
    session_id = "test123"
    
    # Create the .wcgw directory
    wcgw_dir = tmp_path / '.wcgw'
    wcgw_dir.mkdir(parents=True, exist_ok=True)
    
    # Patch Path to return our temp directory
    with patch('pathlib.Path', return_value=wcgw_dir):
        # Create a mock file context
        with patch('builtins.open', mock_open()) as mock_file:
            save_history(history, session_id)
            
            # Verify the file was opened
            mock_file.assert_called_once()
            
            # Get all the written content by joining all write calls
            written_content = ''
            for call_args in mock_file().write.call_args_list:
                written_content += call_args[0][0]
            
            # Verify the content written was valid JSON
            parsed_content = json.loads(written_content)
            assert len(parsed_content) == len(history)
            assert parsed_content == history  # Verify exact content match


def test_parse_user_message_special_text_only():
    msg = "Hello\\nWorld"
    result = parse_user_message_special(msg)
    assert result["role"] == "user"
    assert isinstance(result["content"], list)
    assert len(result["content"]) == 1
    assert result["content"][0]["type"] == "text"
    assert result["content"][0]["text"] == "Hello\\nWorld"


def test_parse_user_message_special_with_image(tmp_path):
    # Create a temporary image file
    image_path = tmp_path / "test.png"
    with open(image_path, "wb") as f:
        f.write(b"fake image data")
    
    msg = f"%image {str(image_path)}\nSome text"
    
    with patch('base64.b64encode', return_value=b"fake_base64"):
        with patch('mimetypes.guess_type', return_value=("image/png", None)):
            result = parse_user_message_special(msg)
            assert result["role"] == "user"
            assert isinstance(result["content"], list)
            assert len(result["content"]) == 2
            assert result["content"][0]["type"] == "image"
            assert result["content"][0]["source"]["type"] == "base64"
            assert result["content"][0]["source"]["media_type"] == "image/png"
            assert result["content"][0]["source"]["data"] == "fake_base64"
            assert result["content"][1]["type"] == "text"
            assert result["content"][1]["text"] == "Some text"


"""Tests for diff_edit.py functionality."""

import unittest

from wcgw.client.file_ops.diff_edit import (
    FileEditOutput,
    Tolerance,
    TolerancesHit,
    find_least_edit_distance_substring,
    match_exact,
    match_with_tolerance,
    match_with_tolerance_empty_line,
    remove_leading_trailing_empty_lines,
)


class TestDiffEdit(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_file_edit_output_replace_or_throw(self):
        # Test successful replacement
        original_content = ["line1", "line2", "line3"]
        search_blocks = [["line2"]]
        replaced = [("line2-new",)]
        edited = [(slice(1, 2, 1), [], ["line2-new"])]

        output = FileEditOutput(
            original_content=original_content,
            orig_search_blocks=search_blocks,
            edited_with_tolerances=edited,
        )

        result, warnings = output.replace_or_throw(max_errors=1)
        self.assertEqual(result, ["line1", "line2-new", "line3"])
        self.assertEqual(warnings, set())

        # Test error cases
        tolerances_hit = TolerancesHit(
            line_process=lambda x: x,
            severity_cat="ERROR",
            score_multiplier=1.0,
            error_name="Test error",
            count=1,
        )
        edited_with_error = [(slice(1, 2, 1), [tolerances_hit], ["line2-new"])]
        output_with_error = FileEditOutput(
            original_content=original_content,
            orig_search_blocks=search_blocks,
            edited_with_tolerances=edited_with_error,
        )

        with self.assertRaises(Exception) as ctx:
            result, warnings = output_with_error.replace_or_throw(max_errors=1)
        self.assertIn("Test error", str(ctx.exception))

    def test_file_edit_output_get_best_match(self):
        # Create multiple outputs with different tolerance hits
        content = ["line1", "line2", "line3"]
        search_blocks = [["line2"]]
        edit1 = FileEditOutput(
            original_content=content,
            orig_search_blocks=search_blocks,
            edited_with_tolerances=[(slice(1, 2, 1), [], ["line2-new"])],
        )

        silent_tolerance = TolerancesHit(
            line_process=lambda x: x, severity_cat="SILENT", score_multiplier=1.0, error_name="", count=1
        )
        edit2 = FileEditOutput(
            original_content=content,
            orig_search_blocks=search_blocks,
            edited_with_tolerances=[
                (slice(1, 2, 1), [silent_tolerance], ["line2-new"])
            ],
        )

        warning_tolerance = TolerancesHit(
            line_process=lambda x: x,
            severity_cat="WARNING",
            score_multiplier=1.0,
            error_name="test warning",
            count=1,
        )
        edit3 = FileEditOutput(
            original_content=content,
            orig_search_blocks=search_blocks,
            edited_with_tolerances=[
                (slice(1, 2, 1), [warning_tolerance], ["line2-new"])
            ],
        )

        # Test preference order: ERROR < WARNING < SILENT
        best_matches, has_error = FileEditOutput.get_best_match([edit1, edit2, edit3])
        self.assertEqual(len(best_matches), 1)
        self.assertEqual(best_matches[0], edit1)
        self.assertFalse(has_error)

    def test_match_exact(self):
        # Test exact matches
        content = ["line1", "line2", "line3", "line2", "line4"]
        search = ["line2"]
        matches = match_exact(content, 0, search)
        self.assertEqual(len(matches), 2)
        self.assertEqual(matches[0], slice(1, 2, 1))
        self.assertEqual(matches[1], slice(3, 4, 1))

        # Test no matches
        search = ["nonexistent"]
        matches = match_exact(content, 0, search)
        self.assertEqual(len(matches), 0)

        # Test empty inputs
        self.assertEqual(len(match_exact([], 0, search)), 0)
        self.assertEqual(len(match_exact(content, 0, [])), 0)

        # Test offset
        matches = match_exact(content, 2, ["line3"])
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0], slice(2, 3, 1))

    def test_match_with_tolerance(self):
        # Test with default tolerances
        content = ["  line1  ", "line2", " line3 "]
        search = ["line1"]
        matches = match_with_tolerance(
            content,
            0,
            search,
            [
                Tolerance(line_process=str.strip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
                Tolerance(line_process=str.lstrip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
                Tolerance(line_process=str.rstrip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
            ],
        )
        self.assertEqual(len(matches), 1)

        # Test multiple matches with different tolerances
        content = ["  line1", "line1  ", " line1 "]
        matches = match_with_tolerance(
            content,
            0,
            search,
            [
                Tolerance(line_process=str.strip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
                Tolerance(line_process=str.lstrip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
                Tolerance(line_process=str.rstrip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
            ],
        )
        self.assertEqual(len(matches), 3)

    def test_match_with_tolerance_empty_line(self):
        # Test with empty lines in content
        content = ["", "line1", "", "line2", "", ""]
        search = ["line1", "line2"]
        matches = match_with_tolerance_empty_line(
            content,
            0,
            search,
            [
                Tolerance(line_process=str.strip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
            ],
        )
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0][0], slice(1, 4, 1))

        # Test with empty lines in search
        content = ["line1", "line2"]
        search = ["", "line1", "", "line2", ""]
        matches = match_with_tolerance_empty_line(
            content,
            0,
            search,
            [
                Tolerance(line_process=str.strip, severity_cat="SILENT", score_multiplier=1.0, error_name=""),
            ],
        )
        self.assertEqual(len(matches), 1)

    def test_find_least_edit_distance_substring(self):
        # Test exact match
        content = ["line1", "line2", "line3"]
        find = ["line2"]
        match, sim, context = find_least_edit_distance_substring(content, 0, find)
        self.assertIsNotNone(match)
        self.assertTrue(sim > 0.9)  # High similarity for exact match

        # Test similar but not exact match
        content = ["line1", "lin2", "line3"]
        find = ["line2"]
        match, sim, context = find_least_edit_distance_substring(content, 0, find)
        self.assertIsNotNone(match)
        self.assertTrue(0.5 < sim < 1.0)  # Moderate similarity

        # Test no good match
        content = ["completely", "different", "content"]
        find = ["line2"]
        match, sim, context = find_least_edit_distance_substring(content, 0, find)
        self.assertTrue(sim < 0.5)  # Low similarity

    def test_remove_leading_trailing_empty_lines(self):
        # Test normal case
        lines = ["", "  ", "content", "more", "", "  "]
        result = remove_leading_trailing_empty_lines(lines)
        self.assertEqual(result, ["content", "more"])

        # Test no empty lines
        lines = ["content", "more"]
        result = remove_leading_trailing_empty_lines(lines)
        self.assertEqual(result, lines)

        # Test all empty lines
        lines = ["", "  ", "", "   "]
        result = remove_leading_trailing_empty_lines(lines)
        self.assertEqual(result, [])

        # Test empty input
        self.assertEqual(remove_leading_trailing_empty_lines([]), [])


if __name__ == "__main__":
    unittest.main()



from .client.session import ClientSession
from .client.stdio import StdioServerParameters, stdio_client
from .server.session import ServerSession
from .server.stdio import stdio_server
from .shared.exceptions import McpError
from .types import (
    CallToolRequest,
    ClientCapabilities,
    ClientNotification,
    ClientRequest,
    ClientResult,
    CompleteRequest,
    CreateMessageRequest,
    CreateMessageResult,
    ErrorData,
    GetPromptRequest,
    GetPromptResult,
    Implementation,
    IncludeContext,
    InitializedNotification,
    InitializeRequest,
    InitializeResult,
    JSONRPCError,
    JSONRPCRequest,
    JSONRPCResponse,
    ListPromptsRequest,
    ListPromptsResult,
    ListResourcesRequest,
    ListResourcesResult,
    ListToolsResult,
    LoggingLevel,
    LoggingMessageNotification,
    Notification,
    PingRequest,
    ProgressNotification,
    PromptsCapability,
    ReadResourceRequest,
    ReadResourceResult,
    Resource,
    ResourcesCapability,
    ResourceUpdatedNotification,
    RootsCapability,
    SamplingMessage,
    ServerCapabilities,
    ServerNotification,
    ServerRequest,
    ServerResult,
    SetLevelRequest,
    StopReason,
    SubscribeRequest,
    Tool,
    ToolsCapability,
    UnsubscribeRequest,
)
from .types import (
    Role as SamplingRole,
)

__all__ = [
    "CallToolRequest",
    "ClientCapabilities",
    "ClientNotification",
    "ClientRequest",
    "ClientResult",
    "ClientSession",
    "CreateMessageRequest",
    "CreateMessageResult",
    "ErrorData",
    "GetPromptRequest",
    "GetPromptResult",
    "Implementation",
    "IncludeContext",
    "InitializeRequest",
    "InitializeResult",
    "InitializedNotification",
    "JSONRPCError",
    "JSONRPCRequest",
    "ListPromptsRequest",
    "ListPromptsResult",
    "ListResourcesRequest",
    "ListResourcesResult",
    "ListToolsResult",
    "LoggingLevel",
    "LoggingMessageNotification",
    "McpError",
    "Notification",
    "PingRequest",
    "ProgressNotification",
    "PromptsCapability",
    "ReadResourceRequest",
    "ReadResourceResult",
    "ResourcesCapability",
    "ResourceUpdatedNotification",
    "Resource",
    "RootsCapability",
    "SamplingMessage",
    "SamplingRole",
    "ServerCapabilities",
    "ServerNotification",
    "ServerRequest",
    "ServerResult",
    "ServerSession",
    "SetLevelRequest",
    "StdioServerParameters",
    "StopReason",
    "SubscribeRequest",
    "Tool",
    "ToolsCapability",
    "UnsubscribeRequest",
    "stdio_client",
    "stdio_server",
    "CompleteRequest",
    "JSONRPCResponse",
]



import warnings
import pytest


@pytest.fixture(autouse=True)
def suppress_forkpty_warning():
    # Suppress forkpty warning
    warnings.filterwarnings("ignore", 
                          category=DeprecationWarning,
                          message="This process .* is multi-threaded")
    yield
    warnings.resetwarnings()



import unittest
from unittest.mock import patch, MagicMock
from wcgw.client.sys_utils import maybe_truncate, command_run, MAX_RESPONSE_LEN, TRUNCATED_MESSAGE
import subprocess
from typing import Tuple


class TestSysUtils(unittest.TestCase):
    def test_maybe_truncate_short_content(self):
        content = "Short content"
        result = maybe_truncate(content)
        self.assertEqual(result, content)

    def test_maybe_truncate_long_content(self):
        content = "x" * (MAX_RESPONSE_LEN + 100)
        result = maybe_truncate(content)
        self.assertEqual(len(result), MAX_RESPONSE_LEN + len(TRUNCATED_MESSAGE))
        self.assertTrue(result.endswith(TRUNCATED_MESSAGE))

    def test_maybe_truncate_custom_length(self):
        content = "Hello World"
        truncate_after = 5
        result = maybe_truncate(content, truncate_after)
        self.assertEqual(result, "Hello" + TRUNCATED_MESSAGE)

    def test_maybe_truncate_none_length(self):
        content = "x" * (MAX_RESPONSE_LEN + 100)
        result = maybe_truncate(content, None)
        self.assertEqual(result, content)

    @patch("subprocess.Popen")
    def test_command_run_success(self, mock_popen):
        # Setup mock
        process_mock = MagicMock()
        process_mock.communicate.return_value = ("stdout", "stderr")
        process_mock.returncode = 0
        mock_popen.return_value = process_mock

        # Run command
        returncode, stdout, stderr = command_run("echo test")

        # Verify results
        self.assertEqual(returncode, 0)
        self.assertEqual(stdout, "stdout")
        self.assertEqual(stderr, "stderr")
        mock_popen.assert_called_once_with(
            "echo test",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )

    @patch("subprocess.Popen")
    def test_command_run_with_timeout(self, mock_popen):
        process_mock = MagicMock()
        process_mock.communicate.side_effect = subprocess.TimeoutExpired("cmd", 1.0)
        process_mock.kill = MagicMock()
        mock_popen.return_value = process_mock

        with self.assertRaises(TimeoutError):
            command_run("sleep 10", timeout=1.0)

        process_mock.kill.assert_called_once()

    @patch("subprocess.Popen")
    def test_command_run_binary_mode(self, mock_popen):
        process_mock = MagicMock()
        process_mock.communicate.return_value = (b"binary stdout", b"binary stderr")
        process_mock.returncode = 0
        mock_popen.return_value = process_mock

        returncode, stdout, stderr = command_run("cat binary_file", text=False)

        self.assertEqual(returncode, 0)
        mock_popen.assert_called_once_with(
            "cat binary_file",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=False
        )

    @patch("subprocess.Popen")
    def test_command_run_truncate_output(self, mock_popen):
        # Create long output that should be truncated
        long_output = "x" * (MAX_RESPONSE_LEN + 100)
        
        process_mock = MagicMock()
        process_mock.communicate.return_value = (long_output, "stderr")
        process_mock.returncode = 0
        mock_popen.return_value = process_mock

        returncode, stdout, stderr = command_run("some command")

        self.assertEqual(returncode, 0)
        self.assertTrue(stdout.endswith(TRUNCATED_MESSAGE))
        self.assertEqual(len(stdout), MAX_RESPONSE_LEN + len(TRUNCATED_MESSAGE))


if __name__ == "__main__":
    unittest.main()



from datetime import timedelta

from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from pydantic import AnyUrl

import mcp_wcgw.types as types
from mcp_wcgw.shared.session import BaseSession
from mcp_wcgw.shared.version import SUPPORTED_PROTOCOL_VERSIONS


class ClientSession(
    BaseSession[
        types.ClientRequest,
        types.ClientNotification,
        types.ClientResult,
        types.ServerRequest,
        types.ServerNotification,
    ]
):
    def __init__(
        self,
        read_stream: MemoryObjectReceiveStream[types.JSONRPCMessage | Exception],
        write_stream: MemoryObjectSendStream[types.JSONRPCMessage],
        read_timeout_seconds: timedelta | None = None,
    ) -> None:
        super().__init__(
            read_stream,
            write_stream,
            types.ServerRequest,
            types.ServerNotification,
            read_timeout_seconds=read_timeout_seconds,
        )

    async def initialize(self) -> types.InitializeResult:
        result = await self.send_request(
            types.ClientRequest(
                types.InitializeRequest(
                    method="initialize",
                    params=types.InitializeRequestParams(
                        protocolVersion=types.LATEST_PROTOCOL_VERSION,
                        capabilities=types.ClientCapabilities(
                            sampling=None,
                            experimental=None,
                            roots=types.RootsCapability(
                                # TODO: Should this be based on whether we
                                # _will_ send notifications, or only whether
                                # they're supported?
                                listChanged=True
                            ),
                        ),
                        clientInfo=types.Implementation(name="mcp", version="0.1.0"),
                    ),
                )
            ),
            types.InitializeResult,
        )

        if result.protocolVersion not in SUPPORTED_PROTOCOL_VERSIONS:
            raise RuntimeError(
                "Unsupported protocol version from the server: "
                f"{result.protocolVersion}"
            )

        await self.send_notification(
            types.ClientNotification(
                types.InitializedNotification(method="notifications/initialized")
            )
        )

        return result

    async def send_ping(self) -> types.EmptyResult:
        """Send a ping request."""
        return await self.send_request(
            types.ClientRequest(
                types.PingRequest(
                    method="ping",
                )
            ),
            types.EmptyResult,
        )

    async def send_progress_notification(
        self, progress_token: str | int, progress: float, total: float | None = None
    ) -> None:
        """Send a progress notification."""
        await self.send_notification(
            types.ClientNotification(
                types.ProgressNotification(
                    method="notifications/progress",
                    params=types.ProgressNotificationParams(
                        progressToken=progress_token,
                        progress=progress,
                        total=total,
                    ),
                ),
            )
        )

    async def set_logging_level(self, level: types.LoggingLevel) -> types.EmptyResult:
        """Send a logging/setLevel request."""
        return await self.send_request(
            types.ClientRequest(
                types.SetLevelRequest(
                    method="logging/setLevel",
                    params=types.SetLevelRequestParams(level=level),
                )
            ),
            types.EmptyResult,
        )

    async def list_resources(self) -> types.ListResourcesResult:
        """Send a resources/list request."""
        return await self.send_request(
            types.ClientRequest(
                types.ListResourcesRequest(
                    method="resources/list",
                )
            ),
            types.ListResourcesResult,
        )

    async def read_resource(self, uri: AnyUrl) -> types.ReadResourceResult:
        """Send a resources/read request."""
        return await self.send_request(
            types.ClientRequest(
                types.ReadResourceRequest(
                    method="resources/read",
                    params=types.ReadResourceRequestParams(uri=uri),
                )
            ),
            types.ReadResourceResult,
        )

    async def subscribe_resource(self, uri: AnyUrl) -> types.EmptyResult:
        """Send a resources/subscribe request."""
        return await self.send_request(
            types.ClientRequest(
                types.SubscribeRequest(
                    method="resources/subscribe",
                    params=types.SubscribeRequestParams(uri=uri),
                )
            ),
            types.EmptyResult,
        )

    async def unsubscribe_resource(self, uri: AnyUrl) -> types.EmptyResult:
        """Send a resources/unsubscribe request."""
        return await self.send_request(
            types.ClientRequest(
                types.UnsubscribeRequest(
                    method="resources/unsubscribe",
                    params=types.UnsubscribeRequestParams(uri=uri),
                )
            ),
            types.EmptyResult,
        )

    async def call_tool(
        self, name: str, arguments: dict | None = None
    ) -> types.CallToolResult:
        """Send a tools/call request."""
        return await self.send_request(
            types.ClientRequest(
                types.CallToolRequest(
                    method="tools/call",
                    params=types.CallToolRequestParams(name=name, arguments=arguments),
                )
            ),
            types.CallToolResult,
        )

    async def list_prompts(self) -> types.ListPromptsResult:
        """Send a prompts/list request."""
        return await self.send_request(
            types.ClientRequest(
                types.ListPromptsRequest(
                    method="prompts/list",
                )
            ),
            types.ListPromptsResult,
        )

    async def get_prompt(
        self, name: str, arguments: dict[str, str] | None = None
    ) -> types.GetPromptResult:
        """Send a prompts/get request."""
        return await self.send_request(
            types.ClientRequest(
                types.GetPromptRequest(
                    method="prompts/get",
                    params=types.GetPromptRequestParams(name=name, arguments=arguments),
                )
            ),
            types.GetPromptResult,
        )

    async def complete(
        self, ref: types.ResourceReference | types.PromptReference, argument: dict
    ) -> types.CompleteResult:
        """Send a completion/complete request."""
        return await self.send_request(
            types.ClientRequest(
                types.CompleteRequest(
                    method="completion/complete",
                    params=types.CompleteRequestParams(
                        ref=ref,
                        argument=types.CompletionArgument(**argument),
                    ),
                )
            ),
            types.CompleteResult,
        )

    async def list_tools(self) -> types.ListToolsResult:
        """Send a tools/list request."""
        return await self.send_request(
            types.ClientRequest(
                types.ListToolsRequest(
                    method="tools/list",
                )
            ),
            types.ListToolsResult,
        )

    async def send_roots_list_changed(self) -> None:
        """Send a roots/list_changed notification."""
        await self.send_notification(
            types.ClientNotification(
                types.RootsListChangedNotification(
                    method="notifications/roots/list_changed",
                )
            )
        )



import pytest
from wcgw.client.tools import _is_int

def test_is_int_validation():
    """Test _is_int function with various inputs"""
    # Valid integers
    assert _is_int("123") is True
    assert _is_int("-123") is True
    assert _is_int("0") is True
    assert _is_int("０１２３") is True  # Full-width numbers are actually valid
    
    # Invalid cases
    assert _is_int("123abc") is False  # Mixed content
    assert _is_int("12.3") is False    # Decimal
    assert _is_int("") is False        # Empty string
    assert _is_int("①②③") is False     # Circle numbers

def test_is_int_edge_cases():
    """Test edge cases for _is_int function"""
    assert _is_int("+123") is True     # Positive sign
    assert _is_int("-0") is True       # Negative zero
    assert _is_int("00123") is True    # Leading zeros
    assert _is_int(" 123") is True     # Leading space
    assert _is_int("123 ") is True     # Trailing space
    assert _is_int(" 123 ") is True    # Both spaces


import json
import os
import re
import shlex
from typing import Any, Callable, Optional

from ..types_ import ContextSave


def get_app_dir_xdg() -> str:
    xdg_data_dir = os.environ.get("XDG_DATA_HOME", os.path.expanduser("~/.local/share"))
    return os.path.join(xdg_data_dir, "wcgw")


def format_memory(task_memory: ContextSave, relevant_files: str) -> str:
    memory_data = ""
    if task_memory.project_root_path:
        memory_data += (
            f"# PROJECT ROOT = {shlex.quote(task_memory.project_root_path)}\n"
        )
    memory_data += task_memory.description

    memory_data += (
        "\n\n"
        + "# Relevant file paths\n"
        + ", ".join(map(shlex.quote, task_memory.relevant_file_globs))
    )

    memory_data += "\n\n# Relevant Files:\n" + relevant_files

    return memory_data


def save_memory(
    task_memory: ContextSave,
    relevant_files: str,
    bash_state_dict: Optional[dict[str, Any]] = None,
) -> str:
    app_dir = get_app_dir_xdg()
    memory_dir = os.path.join(app_dir, "memory")
    os.makedirs(memory_dir, exist_ok=True)

    task_id = task_memory.id
    if not task_id:
        raise Exception("Task id can not be empty")
    memory_data = format_memory(task_memory, relevant_files)

    memory_file_full = os.path.join(memory_dir, f"{task_id}.txt")

    with open(memory_file_full, "w") as f:
        f.write(memory_data)

    # Save bash state if provided
    if bash_state_dict is not None:
        state_file = os.path.join(memory_dir, f"{task_id}_bash_state.json")
        with open(state_file, "w") as f:
            json.dump(bash_state_dict, f, indent=2)

    return memory_file_full


def load_memory[T](
    task_id: str,
    max_tokens: Optional[int],
    encoder: Callable[[str], list[T]],
    decoder: Callable[[list[T]], str],
) -> tuple[str, str, Optional[dict[str, Any]]]:
    app_dir = get_app_dir_xdg()
    memory_dir = os.path.join(app_dir, "memory")
    memory_file = os.path.join(memory_dir, f"{task_id}.txt")

    with open(memory_file, "r") as f:
        data = f.read()

    if max_tokens:
        toks = encoder(data)
        if len(toks) > max_tokens:
            toks = toks[: max(0, max_tokens - 10)]
            data = decoder(toks)
            data += "\n(... truncated)"

    project_root_match = re.search(r"# PROJECT ROOT = \s*(.*?)\s*$", data, re.MULTILINE)
    project_root_path = ""
    if project_root_match:
        matched_path = project_root_match.group(1)
        parsed_ = shlex.split(matched_path)
        if parsed_ and len(parsed_) == 1:
            project_root_path = parsed_[0]

    # Try to load bash state if exists
    state_file = os.path.join(memory_dir, f"{task_id}_bash_state.json")
    bash_state: Optional[dict[str, Any]] = None
    if os.path.exists(state_file):
        with open(state_file) as f:
            bash_state = json.load(f)

    return project_root_path, data, bash_state



import pytest

from mcp_wcgw.client.stdio import StdioServerParameters, stdio_client
from mcp_wcgw.types import JSONRPCMessage, JSONRPCRequest, JSONRPCResponse


@pytest.mark.anyio
async def test_stdio_client():
    server_parameters = StdioServerParameters(command="/usr/bin/tee")

    async with stdio_client(server_parameters) as (read_stream, write_stream):
        # Test sending and receiving messages
        messages = [
            JSONRPCMessage(root=JSONRPCRequest(jsonrpc="2.0", id=1, method="ping")),
            JSONRPCMessage(root=JSONRPCResponse(jsonrpc="2.0", id=2, result={})),
        ]

        async with write_stream:
            for message in messages:
                await write_stream.send(message)

        read_messages = []
        async with read_stream:
            async for message in read_stream:
                if isinstance(message, Exception):
                    raise message

                read_messages.append(message)
                if len(read_messages) == 2:
                    break

        assert len(read_messages) == 2
        assert read_messages[0] == JSONRPCMessage(
            root=JSONRPCRequest(jsonrpc="2.0", id=1, method="ping")
        )
        assert read_messages[1] == JSONRPCMessage(
            root=JSONRPCResponse(jsonrpc="2.0", id=2, result={})
        )



import importlib.metadata
import logging
import sys

import anyio

from mcp_wcgw.server.models import InitializationOptions
from mcp_wcgw.server.session import ServerSession
from mcp_wcgw.server.stdio import stdio_server
from mcp_wcgw.types import ServerCapabilities

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("ignore")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("server")


async def receive_loop(session: ServerSession):
    logger.info("Starting receive loop")
    async for message in session.incoming_messages:
        if isinstance(message, Exception):
            logger.error("Error: %s", message)
            continue

        logger.info("Received message from client: %s", message)


async def main():
    version = importlib.metadata.version("mcp")
    async with stdio_server() as (read_stream, write_stream):
        async with (
            ServerSession(
                read_stream,
                write_stream,
                InitializationOptions(
                    server_name="mcp",
                    server_version=version,
                    capabilities=ServerCapabilities(),
                ),
            ) as session,
            write_stream,
        ):
            await receive_loop(session)


if __name__ == "__main__":
    anyio.run(main, backend="trio")



import base64
import json
import os
import unittest
from unittest.mock import MagicMock, mock_open, patch

from wcgw.client.tools import (
    BASH_STATE,
    ImageData,
    get_incremental_output,
    render_terminal_output,
    which_tool,
    which_tool_name,
)
from wcgw.types_ import BashCommand, BashInteraction, Keyboard, Mouse, WriteIfEmpty


class TestToolsExtended(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        from wcgw.client.tools import BASH_STATE, INITIALIZED, TOOL_CALLS, initialize

        global INITIALIZED, TOOL_CALLS
        INITIALIZED = False
        TOOL_CALLS = []
        BASH_STATE._is_in_docker = ""  # Reset Docker state without shell reset

        # Properly initialize tools for testing
        initialize(
            any_workspace_path="",
            read_files_=[],
            task_id_to_resume="",
            max_tokens=None,
            mode="wcgw",
        )

    def tearDown(self):
        from wcgw.client.tools import BASH_STATE, INITIALIZED, TOOL_CALLS

        global INITIALIZED, TOOL_CALLS
        INITIALIZED = False  # Reset initialization state
        TOOL_CALLS = []  # Clear tool calls
        try:
            BASH_STATE.reset_shell()  # Reset bash state
        except Exception as e:
            print(f"Warning: Failed to reset BASH_STATE: {e}")
        # Clean up any temporary files or directories
        if hasattr(self, "_saved_filepath") and os.path.exists(
            getattr(self, "_saved_filepath")
        ):
            os.remove(self._saved_filepath)

    def test_get_incremental_output(self):
        old_output = ["line1", "line2"]
        new_output = ["line1", "line2", "line3"]

        result = get_incremental_output(old_output, new_output)
        self.assertEqual(result, ["line3"])

        # Test with empty old output
        result = get_incremental_output([], new_output)
        self.assertEqual(result, new_output)

        # Test with completely different output
        result = get_incremental_output(["old"], ["new"])
        self.assertEqual(result, ["new"])

    def test_render_terminal_output(self):
        # Test with ANSI escape sequences
        terminal_output = "\x1b[32mGreen Text\x1b[0m\nNext Line"
        result = render_terminal_output(terminal_output)
        # Strip spaces since terminal width may vary
        result = [line.strip() for line in result]
        self.assertEqual(result, ["Green Text", "Next Line"])

        # Test with carriage returns
        terminal_output = "First\rSecond\nThird"
        result = render_terminal_output(terminal_output)
        self.assertTrue("Second" in result[0])

    def test_which_tool(self):
        # Test BashCommand
        cmd_json = json.dumps({"command": "ls"})
        result = which_tool(cmd_json)
        self.assertIsInstance(result, BashCommand)
        self.assertEqual(result.command, "ls")

        # Test BashInteraction
        interaction_json = json.dumps({"send_text": "input"})
        result = which_tool(interaction_json)
        self.assertIsInstance(result, BashInteraction)
        self.assertEqual(result.send_text, "input")

    def test_which_tool_name(self):
        # Test valid tool names
        self.assertEqual(which_tool_name("BashCommand"), BashCommand)
        self.assertEqual(which_tool_name("Mouse"), Mouse)
        self.assertEqual(which_tool_name("Keyboard"), Keyboard)

        # Test invalid tool name
        with self.assertRaises(ValueError):
            which_tool_name("InvalidTool")

        # Test pending state
        BASH_STATE.set_pending("test output")
        self.assertEqual(BASH_STATE.state, "pending")
        self.assertEqual(BASH_STATE.pending_output, "test output")

        # Test whitelist operations
        BASH_STATE.add_to_whitelist_for_overwrite("/test/path")
        self.assertIn("/test/path", BASH_STATE.whitelist_for_overwrite)

    def test_image_data(self):
        # Test ImageData model
        image = ImageData(media_type="image/png", data="base64data")
        self.assertEqual(image.media_type, "image/png")
        self.assertEqual(image.data, "base64data")
        self.assertEqual(image.dataurl, "data:image/png;base64,base64data")

    @patch("os.path.exists")
    @patch("os.path.isabs")
    @patch("builtins.open", new_callable=mock_open)
    def test_read_image_from_shell(self, mock_file, mock_isabs, mock_exists):
        from wcgw.client.tools import read_image_from_shell

        # Setup mocks
        mock_isabs.return_value = True
        mock_exists.return_value = True
        mock_file.return_value.read.return_value = b"test_image_data"

        # Test regular file read
        result = read_image_from_shell("/test/image.png")
        self.assertIsInstance(result, ImageData)
        self.assertEqual(result.media_type, "image/png")
        self.assertEqual(result.data, base64.b64encode(b"test_image_data").decode())

        # Test non-existent file
        mock_exists.return_value = False
        with self.assertRaises(ValueError):
            read_image_from_shell("/nonexistent/image.png")

    @patch("wcgw.client.tools.default_enc")
    def test_get_context_for_errors(self, mock_enc):
        from wcgw.client.tools import get_context_for_errors

        # Setup mock tokenizer
        mock_enc.encode.return_value = [1, 2, 3]  # simulate tokens

        # Test basic context
        file_content = "line1\nline2\nline3\nline4\nline5"
        errors = [(2, 0)]  # Error on line 2
        result = get_context_for_errors(errors, file_content, max_tokens=100)
        self.assertIn("line2", result)
        self.assertIn("```", result)

        # Test token limit exceeded
        mock_enc.encode.return_value = [i for i in range(200)]  # many tokens
        result = get_context_for_errors(errors, file_content, max_tokens=10)
        self.assertEqual(result, "Please re-read the file to understand the context")

    @patch("os.path.exists")
    @patch("os.path.isabs")
    @patch("pathlib.Path.open")
    @patch("pathlib.Path.mkdir")
    def test_write_file(self, mock_mkdir, mock_path_open, mock_isabs, mock_exists):
        from wcgw.client.tools import BASH_STATE, write_file

        # Setup mocks
        mock_isabs.return_value = True
        mock_exists.return_value = False
        mock_file = mock_open()
        mock_path_open.return_value.__enter__ = mock_file
        mock_path_open.return_value.__exit__ = MagicMock()

        # Test successful write
        test_file = WriteIfEmpty(
            file_path="/test/file.py", file_content="print('test')"
        )
        result = write_file(test_file, error_on_exist=True, max_tokens=100)
        self.assertIn("Success", result)

        # Test writing to existing file with error_on_exist=True and not in whitelist
        mock_exists.return_value = True
        BASH_STATE.whitelist_for_overwrite.clear()  # Clear whitelist
        test_file_new = WriteIfEmpty(
            file_path="/test/another_file.py",  # Use a different file not in whitelist
            file_content="print('test')",
        )
        with patch("pathlib.Path.read_text") as mock_read_text:
            mock_read_text.return_value = "existing content"
            result = write_file(test_file_new, error_on_exist=True, max_tokens=100)
            self.assertIn("Error: can't write to existing file", result)

        # Test with relative path
        mock_isabs.return_value = False
        result = write_file(test_file, error_on_exist=True, max_tokens=100)
        self.assertIn("Failure: file_path should be absolute path", result)

    def test_is_status_check(self):
        from wcgw.client.tools import is_status_check

        # Test with Enter special key
        interaction = BashInteraction(send_specials=["Enter"])
        self.assertTrue(is_status_check(interaction))

        # Test with ascii code 10 (newline)
        interaction = BashInteraction(send_ascii=[10])
        self.assertTrue(is_status_check(interaction))

        # Test with other interaction
        interaction = BashInteraction(send_text="hello")
        self.assertFalse(is_status_check(interaction))

        # Test with BashCommand
        cmd = BashCommand(command="ls")
        self.assertFalse(is_status_check(cmd))

    @patch("pexpect.spawn")
    def test_start_shell(self, mock_spawn):
        from wcgw.client.tools import PROMPT, start_shell

        # Setup mock shell
        mock_shell = MagicMock()
        mock_spawn.return_value = mock_shell

        # Test successful shell start
        shell = start_shell(is_restricted_mode=False, initial_dir="/")
        self.assertEqual(shell, mock_shell)

        # Verify shell initialization
        self.assertEqual(mock_shell.expect.call_count, 4)  # 4 setup commands
        mock_shell.sendline.assert_any_call(f"export PS1={PROMPT}")
        mock_shell.sendline.assert_any_call("stty -icanon -echo")
        mock_shell.sendline.assert_any_call("set +o pipefail")
        mock_shell.sendline.assert_any_call("export GIT_PAGER=cat PAGER=cat")

        # Test restricted mode
        mock_shell.reset_mock()
        shell = start_shell(is_restricted_mode=True, initial_dir="/")
        self.assertEqual(shell, mock_shell)

    def test_save_out_of_context(self):
        from wcgw.client.tools import save_out_of_context

        # Test saving content
        content = "Test content"
        suffix = ".txt"
        filepath = save_out_of_context(content, suffix)

        # Verify file was created and content saved
        self.assertTrue(os.path.exists(filepath))
        with open(filepath, "r") as f:
            saved_content = f.read()
            self.assertEqual(saved_content, content)

        # Cleanup
        os.remove(filepath)

    @patch("wcgw.client.tools.get_tool_output")
    def test_which_tool_errors(self, mock_get_tool_output):
        from wcgw.client.tools import which_tool

        # Test with invalid JSON
        with self.assertRaises(json.JSONDecodeError):
            which_tool("invalid json")

    @patch("os.system")
    @patch("tempfile.TemporaryDirectory")
    def test_write_file_docker(self, mock_temp_dir, mock_system):
        from wcgw.client.tools import BASH_STATE, write_file

        # Setup Docker environment
        BASH_STATE.set_in_docker("test_container")

        # Setup mocks
        mock_temp_dir.return_value.__enter__.return_value = "/tmp/test"
        mock_system.return_value = 0

        # Test writing in Docker environment
        test_file = WriteIfEmpty(
            file_path="/test/file.py", file_content="print('test')"
        )
        result = write_file(test_file, error_on_exist=False, max_tokens=100)
        self.assertIn("Success", result)

        # Test Docker command failure
        mock_system.return_value = 1
        result = write_file(test_file, error_on_exist=False, max_tokens=100)
        self.assertIn("Error: Write failed with code", result)

    @patch("wcgw.client.tools.command_run")
    def test_read_files_docker(self, mock_command_run):
        from wcgw.client.tools import BASH_STATE, DisableConsole, read_files

        # Setup mocks and test environment
        console = DisableConsole()
        BASH_STATE.set_in_docker("test_container")
        mock_command_run.return_value = (0, "file content", "")

        with patch("wcgw.client.tools.read_file") as mock_read_file:
            mock_read_file.return_value = ("file content", False, 10)

            # Test file read
            result = read_files(["/test/file.py"], max_tokens=100)
            self.assertIn("file content", result)

            # Cleanup
            BASH_STATE._is_in_docker = ""

        # Test read failure
        mock_command_run.return_value = (1, "", "error message")
        with patch("os.path.exists", return_value=False):
            result = read_files(["/test/nonexistent.py"], max_tokens=100)
        self.assertIn("file /test/nonexistent.py does not exist", result)

        # Reset Docker state
        BASH_STATE._is_in_docker = ""

    @patch("wcgw.client.tools.get_tool_output")
    def test_execute_bash_interaction(self, mock_get_tool):
        from wcgw.client.tools import BashInteraction, execute_bash

        mock_tokenizer = MagicMock()
        mock_tokenizer.encode.return_value.ids = [1, 2, 3]

        # Test sending special keys
        interaction = BashInteraction(
            send_specials=[
                "Enter",
                "Key-up",
                "Key-down",
                "Key-left",
                "Key-right",
                "Ctrl-c",
                "Ctrl-d",
                "Ctrl-z",
            ],
        )
        result, _ = execute_bash(
            mock_tokenizer, interaction, max_tokens=100, timeout_s=1
        )
        self.assertIsInstance(result, str)

        # Test sending ASCII characters
        interaction = BashInteraction(
            send_ascii=[97, 98, 99],  # 'abc'
        )
        result, _ = execute_bash(
            mock_tokenizer, interaction, max_tokens=100, timeout_s=1
        )
        self.assertIsInstance(result, str)

        # Test malformed interaction
        interaction = BashInteraction(
            send_text=None, send_ascii=None, send_specials=None
        )
        result, _ = execute_bash(
            mock_tokenizer, interaction, max_tokens=100, timeout_s=1
        )
        self.assertIn("Failure", result)

    @patch("wcgw.client.tools.check_syntax")
    @patch("os.system")
    def test_write_file_with_syntax_check(self, mock_system, mock_check):
        from wcgw.client.tools import write_file

        # Setup mocks
        mock_error = MagicMock()
        mock_error.description = "Invalid syntax"
        mock_error.errors = [(1, 0)]
        mock_check.return_value = mock_error
        mock_system.return_value = 0

        # Test file write with syntax error
        test_file = WriteIfEmpty(
            file_path="/test/file.py", file_content="invalid python code"
        )

        with patch("pathlib.Path.open", mock_open()):
            with patch("pathlib.Path.mkdir"):
                with patch("os.path.exists", return_value=False):
                    with patch("os.path.isabs", return_value=True):
                        result = write_file(
                            test_file, error_on_exist=True, max_tokens=100
                        )
                        self.assertIn("Success", result)
                        self.assertIn("syntax errors", result)
                        self.assertIn("Invalid syntax", result)

    @patch("wcgw.client.tools.read_image_from_shell")
    @patch("wcgw.client.tools.execute_bash")
    @patch("pathlib.Path.mkdir")
    @patch("pathlib.Path.open", new_callable=mock_open)
    def test_get_tool_output_file_operations(
        self, mock_file, mock_mkdir, mock_execute_bash, mock_read_image
    ):
        """Test get_tool_output function with file operation tools"""
        from wcgw.client.tools import get_tool_output

        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Test ReadImage tool
        mock_read_image.return_value = ImageData(
            media_type="image/png", data="test_data"
        )
        result, cost = get_tool_output(
            {"file_path": "/test/image.png"},
            mock_enc,
            1.0,
            mock_loop_call,
            100,
        )
        self.assertIsInstance(result[0], ImageData)
        self.assertEqual(result[0].media_type, "image/png")

        # Test WriteIfEmpty tool
        result, cost = get_tool_output(
            {
                "file_path": "/test/file.txt",
                "file_content": "test content",
            },
            mock_enc,
            1.0,
            mock_loop_call,
            100,
        )
        self.assertTrue(isinstance(result[0], str))

    @patch("wcgw.client.tools.run_computer_tool")
    def test_get_tool_output_computer_interactions(self, mock_run_computer):
        """Test get_tool_output function with computer interaction tools"""
        from wcgw.client.tools import BASH_STATE, get_tool_output

        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Setup mock return value for computer tool
        mock_run_computer.return_value = ("Tool output", "screenshot_data")

        # Test Mouse tool
        result, cost = get_tool_output(
            {
                "action": {"button_type": "left_click"},
            },
            mock_enc,
            1.0,
            mock_loop_call,
            100,
        )
        self.assertEqual(len(result), 2)  # Output string and screenshot
        self.assertTrue(isinstance(result[0], str))

        # Test Keyboard tool with GetScreenInfo
        BASH_STATE.set_in_docker("test_container")
        result, cost = get_tool_output(
            {"action": "type", "text": "test input"},
            mock_enc,
            1.0,
            mock_loop_call,
            100,
        )
        self.assertEqual(len(result), 2)
        BASH_STATE._is_in_docker = ""

    @patch("wcgw.client.tools.take_help_of_ai_assistant")
    def test_get_tool_output_ai_assistant(self, mock_ai_helper):
        """Test get_tool_output function with AI Assistant tool"""
        from wcgw.client.tools import get_tool_output

        mock_enc = MagicMock()
        mock_loop_call = MagicMock()
        mock_ai_helper.return_value = ("AI response", 0.1)

        # Test AIAssistant tool
        result, cost = get_tool_output(
            {
                "instruction": "test instruction",
                "desired_output": "test output",
            },
            mock_enc,
            1.0,
            mock_loop_call,
            100,
        )
        self.assertEqual(result[0], "AI response")
        self.assertEqual(cost, 0.1)

    def test_get_tool_output_invalid_tool(self):
        """Test get_tool_output function with invalid tool"""
        from wcgw.client.tools import get_tool_output

        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Test with None
        with self.assertRaises(ValueError) as cm:
            get_tool_output(None, mock_enc, 1.0, mock_loop_call, 100)
        self.assertEqual(str(cm.exception), "Unknown tool: None")

        # Test with invalid tool type
        with self.assertRaises(ValueError) as cm:
            get_tool_output(123, mock_enc, 1.0, mock_loop_call, 100)

        # Test with empty dict
        result, cost = get_tool_output({}, mock_enc, 1.0, mock_loop_call, 100)
        self.assertIn("Failure:", result[0])
        self.assertEqual(cost, 0)

    def test_get_tool_output_exception_handling(self):
        """Test error handling in get_tool_output"""
        from wcgw.client.tools import get_tool_output

        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Create a write tool for testing with a relative path
        # This should raise a validation error without needing to mock write_file
        write_tool = WriteIfEmpty(file_path="relative/path", file_content="test")

        # Test: function should catch the validation error and return an error message
        result, cost = get_tool_output(write_tool, mock_enc, 1.0, mock_loop_call, 100)

        # Verify the error is handled gracefully
        self.assertEqual(cost, 0)  # Cost should be 0 when there's an error
        self.assertTrue(isinstance(result, list))  # Result should be a list
        self.assertEqual(len(result), 1)  # Should have one item
        self.assertTrue(isinstance(result[0], str))  # Should be a string message
        # Error message should mention the path issue
        self.assertIn("file_path should be absolute path", result[0])


if __name__ == "__main__":
    unittest.main()



import subprocess

MAX_RESPONSE_LEN: int = 16000
TRUNCATED_MESSAGE: str = "<response clipped><NOTE>To save on context only part of this file has been shown to you.</NOTE>"


def maybe_truncate(content: str, truncate_after: int | None = MAX_RESPONSE_LEN) -> str:
    """Truncate content and append a notice if content exceeds the specified length."""
    return (
        content
        if not truncate_after or len(content) <= truncate_after
        else content[:truncate_after] + TRUNCATED_MESSAGE
    )


def command_run(
    cmd: str,
    timeout: float | None = 3.0,  # seconds
    truncate_after: int | None = MAX_RESPONSE_LEN,
    text: bool = True,
) -> tuple[int, str, str]:
    """Run a shell command synchronously with a timeout."""
    try:
        process = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=text,
        )
        stdout, stderr = process.communicate(timeout=timeout)
        return (
            process.returncode or 0,
            maybe_truncate(stdout, truncate_after=truncate_after),
            maybe_truncate(stderr, truncate_after=truncate_after),
        )
    except subprocess.TimeoutExpired as exc:
        process.kill()
        raise TimeoutError(
            f"Command '{cmd}' timed out after {timeout} seconds"
        ) from exc



import argparse
import logging
import sys
from functools import partial
from urllib.parse import urlparse

import anyio

from mcp_wcgw.client.session import ClientSession
from mcp_wcgw.client.sse import sse_client
from mcp_wcgw.client.stdio import StdioServerParameters, stdio_client

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("ignore")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("client")


async def receive_loop(session: ClientSession):
    logger.info("Starting receive loop")
    async for message in session.incoming_messages:
        if isinstance(message, Exception):
            logger.error("Error: %s", message)
            continue

        logger.info("Received message from server: %s", message)


async def run_session(read_stream, write_stream):
    async with (
        ClientSession(read_stream, write_stream) as session,
        anyio.create_task_group() as tg,
    ):
        tg.start_soon(receive_loop, session)

        logger.info("Initializing session")
        await session.initialize()
        logger.info("Initialized")


async def main(command_or_url: str, args: list[str], env: list[tuple[str, str]]):
    env_dict = dict(env)

    if urlparse(command_or_url).scheme in ("http", "https"):
        # Use SSE client for HTTP(S) URLs
        async with sse_client(command_or_url) as streams:
            await run_session(*streams)
    else:
        # Use stdio client for commands
        server_parameters = StdioServerParameters(
            command=command_or_url, args=args, env=env_dict
        )
        async with stdio_client(server_parameters) as streams:
            await run_session(*streams)


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("command_or_url", help="Command or URL to connect to")
    parser.add_argument("args", nargs="*", help="Additional arguments")
    parser.add_argument(
        "-e",
        "--env",
        nargs=2,
        action="append",
        metavar=("KEY", "VALUE"),
        help="Environment variables to set. Can be used multiple times.",
        default=[],
    )

    args = parser.parse_args()
    anyio.run(partial(main, args.command_or_url, args.args, args.env), backend="trio")


if __name__ == "__main__":
    cli()



import json
import unittest
from unittest.mock import MagicMock, patch

from wcgw.client.tools import (
    BASH_STATE,
    ensure_no_previous_output,
    get_tool_output,
    truncate_if_over,
    which_tool,
    which_tool_name,
)
from wcgw.types_ import (
    BashCommand,
    BashInteraction,
    Keyboard,
    Mouse,
    WriteIfEmpty,
)


class TestToolsValidation(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        from wcgw.client.tools import BASH_STATE, initialize

        BASH_STATE.reset_shell()
        # Properly initialize tools for testing
        initialize(
            any_workspace_path="",
            read_files_=[],
            task_id_to_resume="",
            max_tokens=None,
            mode="wcgw",
        )

    def tearDown(self):
        from wcgw.client.tools import BASH_STATE, INITIALIZED, TOOL_CALLS

        global INITIALIZED, TOOL_CALLS
        INITIALIZED = False  # Reset initialization state
        TOOL_CALLS = []  # Clear tool calls
        try:
            BASH_STATE.reset_shell()  # Reset bash state
        except:
            pass

    def test_ensure_no_previous_output_decorator(self):
        """Test ensure_no_previous_output decorator"""

        @ensure_no_previous_output
        def test_func():
            return "success"

        # Test when no previous output
        BASH_STATE.set_repl()
        result = test_func()
        self.assertEqual(result, "success")

        # Test when previous output exists
        BASH_STATE.set_pending("previous output")
        with self.assertRaises(ValueError) as context:
            test_func()
        self.assertIn("A command is already running", str(context.exception))

    def test_truncate_if_over(self):
        """Test truncate_if_over function"""
        mock_enc = MagicMock()
        mock_enc.encode = MagicMock()
        mock_enc.decode = MagicMock()

        with patch("wcgw.client.tools.default_enc", mock_enc):
            # Test with content under limit
            content = "short content"
            mock_enc.encode.return_value = MagicMock(ids=list(range(5)))  # Under limit
            result = truncate_if_over(content, max_tokens=200)
            self.assertEqual(
                result, content
            )  # Should return original content when under limit
            self.assertEqual(
                mock_enc.decode.call_count, 0
            )  # Decode shouldn't be called

            # Test with content over limit
            long_content = "very long content" * 100
            mock_encoding = MagicMock()
            mock_encoding.ids = list(range(200))  # Over limit
            mock_encoding.__len__.return_value = 200  # Make len(tokens) return 200
            mock_enc.encode.return_value = mock_encoding
            mock_enc.decode.return_value = "truncated content"

            result = truncate_if_over(long_content, max_tokens=50)
            # In truncate_if_over: max(0, max_tokens - 100) = max(0, 50-100)
            truncated_ids = []  # Since 50-100 = -50, max(0, -50) = 0
            mock_enc.decode.assert_called_once_with(truncated_ids)
            self.assertEqual(result, "truncated content\n(...truncated)")

            # Test with None max_tokens
            result = truncate_if_over(long_content, max_tokens=None)
            self.assertEqual(result, long_content)

            # Test with no token limit
            result = truncate_if_over(long_content, max_tokens=None)
            self.assertEqual(result, long_content)

    def test_which_tool(self):
        """Test which_tool function"""
        # Test BashCommand
        cmd_json = json.dumps({"command": "ls"})
        result = which_tool(cmd_json)
        self.assertIsInstance(result, BashCommand)
        self.assertEqual(result.command, "ls")

        # Test BashInteraction
        interaction_json = json.dumps({"send_text": "input"})
        result = which_tool(interaction_json)
        self.assertIsInstance(result, BashInteraction)
        self.assertEqual(result.send_text, "input")

        # Test invalid JSON
        with self.assertRaises(json.JSONDecodeError):
            which_tool("invalid json")

        # Test invalid tool type
        invalid_json = json.dumps({"type": "InvalidTool"})
        with self.assertRaises(ValueError):
            which_tool(invalid_json)

    def test_which_tool_name(self):
        """Test which_tool_name function"""
        # Test valid tool names
        self.assertEqual(which_tool_name("BashCommand"), BashCommand)
        self.assertEqual(which_tool_name("BashInteraction"), BashInteraction)
        self.assertEqual(which_tool_name("WriteIfEmpty"), WriteIfEmpty)
        self.assertEqual(which_tool_name("Mouse"), Mouse)
        self.assertEqual(which_tool_name("Keyboard"), Keyboard)

        # Test invalid tool name
        with self.assertRaises(ValueError):
            which_tool_name("InvalidTool")

    @patch("wcgw.client.tools.execute_bash")
    def test_get_tool_output(self, mock_execute_bash):
        """Test get_tool_output function"""
        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Test BashCommand
        mock_execute_bash.return_value = ("command output", 0)
        result, cost = get_tool_output(
            BashCommand(command="ls"), mock_enc, 1.0, mock_loop_call, 100
        )
        self.assertEqual(result[0], "command output")
        self.assertEqual(cost, 0)

        # Test WriteIfEmpty
        test_file = WriteIfEmpty(file_path="/test/file.txt", file_content="test")
        with patch("wcgw.client.tools.write_file") as mock_write:
            mock_write.return_value = "Success"
            result, cost = get_tool_output(
                test_file, mock_enc, 1.0, mock_loop_call, 100
            )
            self.assertEqual(result[0], "Success")
            self.assertEqual(cost, 0)

        # Test error handling
        with self.assertRaises(ValueError):
            get_tool_output({"type": "InvalidTool"}, mock_enc, 1.0, mock_loop_call, 100)

    def test_get_tool_output_exceptions(self):
        """Test get_tool_output exception handling"""
        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Test with relative path in WriteIfEmpty
        test_file = WriteIfEmpty(file_path="relative/path.txt", file_content="test")
        result, cost = get_tool_output(test_file, mock_enc, 1.0, mock_loop_call, 100)
        self.assertIn("file_path should be absolute path", result[0])
        self.assertEqual(cost, 0)

        # Test with invalid tool type
        with self.assertRaises(ValueError) as context:
            get_tool_output({"type": "InvalidTool"}, mock_enc, 1.0, mock_loop_call, 100)
        self.assertIn("validation errors for union", str(context.exception))

    def test_get_tool_output_computer_tools(self):
        """Test get_tool_output with computer interaction tools"""
        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Test Mouse tool
        with patch("wcgw.client.tools.run_computer_tool") as mock_run:
            mock_run.return_value = ("Mouse clicked", "screenshot_data")
            result, cost = get_tool_output(
                Mouse(action={"button_type": "left_click"}),
                mock_enc,
                1.0,
                mock_loop_call,
                100,
            )
            self.assertEqual(result[0], "Mouse clicked")
            self.assertTrue(hasattr(result[1], "media_type"))
            self.assertEqual(result[1].media_type, "image/png")
            self.assertEqual(result[1].data, "screenshot_data")

        # Test Keyboard tool
        with patch("wcgw.client.tools.run_computer_tool") as mock_run:
            mock_run.return_value = ("Keys typed", "screenshot_data")
            result, cost = get_tool_output(
                Keyboard(action="type", text="test"), mock_enc, 1.0, mock_loop_call, 100
            )
            self.assertEqual(result[0], "Keys typed")
            self.assertTrue(hasattr(result[1], "media_type"))
            self.assertEqual(result[1].media_type, "image/png")
            self.assertEqual(result[1].data, "screenshot_data")

    @patch("wcgw.client.tools.run_computer_tool")
    def test_get_tool_output_error_propagation(self, mock_run):
        """Test error propagation in get_tool_output"""
        mock_enc = MagicMock()
        mock_loop_call = MagicMock()

        # Test error propagation from computer tools
        mock_run.return_value = (
            "Error: Computer tool error",
            None,
        )  # Return error style tuple instead of raising
        result, cost = get_tool_output(
            Mouse(action={"button_type": "left_click"}),
            mock_enc,
            1.0,
            mock_loop_call,
            100,
        )
        # When error happens, first element should be error string, second element should be None
        self.assertEqual(result[0], "Error: Computer tool error")
        self.assertEqual(cost, 0)

        # Test error propagation from bash execution
        with patch("wcgw.client.tools.execute_bash") as mock_bash:
            mock_bash.return_value = (
                "Error: Bash error",
                0,
            )  # Return error as string, not as exception
            result, cost = get_tool_output(
                BashCommand(command="ls"), mock_enc, 1.0, mock_loop_call, 100
            )
            self.assertEqual(result[0], "Error: Bash error")
            self.assertEqual(cost, 0)
            self.assertIn("Bash error", result[0])
            self.assertEqual(cost, 0)



from contextlib import AbstractAsyncContextManager
from datetime import timedelta
from typing import Generic, TypeVar

import anyio
import anyio.lowlevel
import httpx
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from pydantic import BaseModel

from mcp_wcgw.shared.exceptions import McpError
from mcp_wcgw.types import (
    ClientNotification,
    ClientRequest,
    ClientResult,
    ErrorData,
    JSONRPCError,
    JSONRPCMessage,
    JSONRPCNotification,
    JSONRPCRequest,
    JSONRPCResponse,
    RequestParams,
    ServerNotification,
    ServerRequest,
    ServerResult,
)

SendRequestT = TypeVar("SendRequestT", ClientRequest, ServerRequest)
SendResultT = TypeVar("SendResultT", ClientResult, ServerResult)
SendNotificationT = TypeVar("SendNotificationT", ClientNotification, ServerNotification)
ReceiveRequestT = TypeVar("ReceiveRequestT", ClientRequest, ServerRequest)
ReceiveResultT = TypeVar("ReceiveResultT", bound=BaseModel)
ReceiveNotificationT = TypeVar(
    "ReceiveNotificationT", ClientNotification, ServerNotification
)

RequestId = str | int


class RequestResponder(Generic[ReceiveRequestT, SendResultT]):
    def __init__(
        self,
        request_id: RequestId,
        request_meta: RequestParams.Meta | None,
        request: ReceiveRequestT,
        session: "BaseSession",
    ) -> None:
        self.request_id = request_id
        self.request_meta = request_meta
        self.request = request
        self._session = session
        self._responded = False

    async def respond(self, response: SendResultT | ErrorData) -> None:
        assert not self._responded, "Request already responded to"
        self._responded = True

        await self._session._send_response(
            request_id=self.request_id, response=response
        )


class BaseSession(
    AbstractAsyncContextManager,
    Generic[
        SendRequestT,
        SendNotificationT,
        SendResultT,
        ReceiveRequestT,
        ReceiveNotificationT,
    ],
):
    """
    Implements an MCP "session" on top of read/write streams, including features
    like request/response linking, notifications, and progress.

    This class is an async context manager that automatically starts processing
    messages when entered.
    """

    _response_streams: dict[
        RequestId, MemoryObjectSendStream[JSONRPCResponse | JSONRPCError]
    ]
    _request_id: int

    def __init__(
        self,
        read_stream: MemoryObjectReceiveStream[JSONRPCMessage | Exception],
        write_stream: MemoryObjectSendStream[JSONRPCMessage],
        receive_request_type: type[ReceiveRequestT],
        receive_notification_type: type[ReceiveNotificationT],
        # If none, reading will never time out
        read_timeout_seconds: timedelta | None = None,
    ) -> None:
        self._read_stream = read_stream
        self._write_stream = write_stream
        self._response_streams = {}
        self._request_id = 0
        self._receive_request_type = receive_request_type
        self._receive_notification_type = receive_notification_type
        self._read_timeout_seconds = read_timeout_seconds

        self._incoming_message_stream_writer, self._incoming_message_stream_reader = (
            anyio.create_memory_object_stream[
                RequestResponder[ReceiveRequestT, SendResultT]
                | ReceiveNotificationT
                | Exception
            ]()
        )

    async def __aenter__(self):
        self._task_group = anyio.create_task_group()
        await self._task_group.__aenter__()
        self._task_group.start_soon(self._receive_loop)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        # Using BaseSession as a context manager should not block on exit (this
        # would be very surprising behavior), so make sure to cancel the tasks
        # in the task group.
        self._task_group.cancel_scope.cancel()
        return await self._task_group.__aexit__(exc_type, exc_val, exc_tb)

    async def send_request(
        self,
        request: SendRequestT,
        result_type: type[ReceiveResultT],
    ) -> ReceiveResultT:
        """
        Sends a request and wait for a response. Raises an McpError if the
        response contains an error.

        Do not use this method to emit notifications! Use send_notification()
        instead.
        """

        request_id = self._request_id
        self._request_id = request_id + 1

        response_stream, response_stream_reader = anyio.create_memory_object_stream[
            JSONRPCResponse | JSONRPCError
        ](1)
        self._response_streams[request_id] = response_stream

        jsonrpc_request = JSONRPCRequest(
            jsonrpc="2.0",
            id=request_id,
            **request.model_dump(by_alias=True, mode="json", exclude_none=True),
        )

        # TODO: Support progress callbacks

        await self._write_stream.send(JSONRPCMessage(jsonrpc_request))

        try:
            with anyio.fail_after(
                None
                if self._read_timeout_seconds is None
                else self._read_timeout_seconds.total_seconds()
            ):
                response_or_error = await response_stream_reader.receive()
        except TimeoutError:
            raise McpError(
                ErrorData(
                    code=httpx.codes.REQUEST_TIMEOUT,
                    message=(
                        f"Timed out while waiting for response to "
                        f"{request.__class__.__name__}. Waited "
                        f"{self._read_timeout_seconds} seconds."
                    ),
                )
            )

        if isinstance(response_or_error, JSONRPCError):
            raise McpError(response_or_error.error)
        else:
            return result_type.model_validate(response_or_error.result)

    async def send_notification(self, notification: SendNotificationT) -> None:
        """
        Emits a notification, which is a one-way message that does not expect
        a response.
        """
        jsonrpc_notification = JSONRPCNotification(
            jsonrpc="2.0",
            **notification.model_dump(by_alias=True, mode="json", exclude_none=True),
        )

        await self._write_stream.send(JSONRPCMessage(jsonrpc_notification))

    async def _send_response(
        self, request_id: RequestId, response: SendResultT | ErrorData
    ) -> None:
        if isinstance(response, ErrorData):
            jsonrpc_error = JSONRPCError(jsonrpc="2.0", id=request_id, error=response)
            await self._write_stream.send(JSONRPCMessage(jsonrpc_error))
        else:
            jsonrpc_response = JSONRPCResponse(
                jsonrpc="2.0",
                id=request_id,
                result=response.model_dump(
                    by_alias=True, mode="json", exclude_none=True
                ),
            )
            await self._write_stream.send(JSONRPCMessage(jsonrpc_response))

    async def _receive_loop(self) -> None:
        async with (
            self._read_stream,
            self._write_stream,
            self._incoming_message_stream_writer,
        ):
            async for message in self._read_stream:
                if isinstance(message, Exception):
                    await self._incoming_message_stream_writer.send(message)
                elif isinstance(message.root, JSONRPCRequest):
                    validated_request = self._receive_request_type.model_validate(
                        message.root.model_dump(
                            by_alias=True, mode="json", exclude_none=True
                        )
                    )
                    responder = RequestResponder(
                        request_id=message.root.id,
                        request_meta=validated_request.root.params._meta
                        if validated_request.root.params
                        else None,
                        request=validated_request,
                        session=self,
                    )

                    await self._received_request(responder)
                    if not responder._responded:
                        await self._incoming_message_stream_writer.send(responder)
                elif isinstance(message.root, JSONRPCNotification):
                    notification = self._receive_notification_type.model_validate(
                        message.root.model_dump(
                            by_alias=True, mode="json", exclude_none=True
                        )
                    )

                    await self._received_notification(notification)
                    await self._incoming_message_stream_writer.send(notification)
                else:  # Response or error
                    stream = self._response_streams.pop(message.root.id, None)
                    if stream:
                        await stream.send(message.root)
                    else:
                        await self._incoming_message_stream_writer.send(
                            RuntimeError(
                                "Received response with an unknown "
                                f"request ID: {message}"
                            )
                        )

    async def _received_request(
        self, responder: RequestResponder[ReceiveRequestT, SendResultT]
    ) -> None:
        """
        Can be overridden by subclasses to handle a request without needing to
        listen on the message stream.

        If the request is responded to within this method, it will not be
        forwarded on to the message stream.
        """

    async def _received_notification(self, notification: ReceiveNotificationT) -> None:
        """
        Can be overridden by subclasses to handle a notification without needing
        to listen on the message stream.
        """

    async def send_progress_notification(
        self, progress_token: str | int, progress: float, total: float | None = None
    ) -> None:
        """
        Sends a progress notification for a request that is currently being
        processed.
        """

    @property
    def incoming_messages(
        self,
    ) -> MemoryObjectReceiveStream[
        RequestResponder[ReceiveRequestT, SendResultT]
        | ReceiveNotificationT
        | Exception
    ]:
        return self._incoming_message_stream_reader



import json
import os
import unittest
from unittest.mock import mock_open, patch

from wcgw.client.memory import format_memory, get_app_dir_xdg, load_memory, save_memory
from wcgw.types_ import ContextSave


class TestMemory(unittest.TestCase):
    class MockTokens:
        def __init__(self, ids):
            self.ids = ids

        def __len__(self):
            return len(self.ids)

        def __getitem__(self, key):
            if isinstance(key, slice):
                return TestMemory.MockTokens(self.ids[key])
            return self.ids[key]

    def test_get_app_dir_xdg(self):
        with patch.dict("os.environ", {"XDG_DATA_HOME": "/custom/data"}):
            result = get_app_dir_xdg()
            self.assertEqual(result, os.path.join("/custom/data", "wcgw"))

        with patch.dict("os.environ", {}, clear=True):
            with patch("os.path.expanduser") as mock_expanduser:
                mock_expanduser.return_value = "/home/user"
                result = get_app_dir_xdg()
                self.assertEqual(result, "/home/user/wcgw")

    def test_format_memory(self):
        task_memory = ContextSave(
            id="test-id",
            project_root_path="/project",
            description="Test description",
            relevant_file_globs=["*.py"],
        )
        relevant_files = "file1.py\nfile2.py"
        result = format_memory(task_memory, relevant_files)
        self.assertIn("# PROJECT ROOT = /project", result)
        self.assertIn("Test description", result)
        self.assertIn("*.py", result)
        self.assertIn("file1.py", result)

    def test_save_memory(self):
        task_memory = ContextSave(
            id="test-id",
            project_root_path="/project",
            description="Test description",
            relevant_file_globs=["*.py"],
        )
        relevant_files = "file1.py\nfile2.py"

        with patch("os.makedirs"), patch("builtins.open", mock_open()) as mock_file:
            result = save_memory(task_memory, relevant_files)
            self.assertIn("test-id.txt", result)
            mock_file().write.assert_called_once()

    def test_load_memory(self):
        """Test loading of memory data including bash state"""
        task_id = "test-id"
        memory_data = "# PROJECT ROOT = /project\nTest description\n*.py\nfile1.py"
        bash_state_data = {
            "bash_command_mode": {
                "bash_mode": "normal_mode",
                "allowed_commands": "all",
            },
            "file_edit_mode": {"allowed_globs": "all"},
            "write_if_empty_mode": {"allowed_globs": "all"},
            "whitelist_for_overwrite": [],
            "mode": "wcgw",
        }
        mock_encoder = lambda x: self.MockTokens([1, 2, 3])  # Simulate tokenizer output
        mock_decoder = lambda x: "Decoded text"  # Return fixed text for any input IDs

        # Mock both the memory file and bash state file
        from wcgw.client.memory import get_app_dir_xdg

        app_dir = get_app_dir_xdg()
        memory_dir = os.path.join(app_dir, "memory")
        mock_files = {
            os.path.join(memory_dir, f"{task_id}.txt"): memory_data,
            os.path.join(memory_dir, f"{task_id}_bash_state.json"): json.dumps(
                bash_state_data
            ),
        }

        def mock_open_file(filename, *args, **kwargs):
            content = mock_files.get(filename)
            if content is None:
                raise FileNotFoundError(filename)
            return mock_open(read_data=content)()

        with patch("builtins.open", side_effect=mock_open_file), \
             patch("os.path.exists", lambda x: x.endswith("_bash_state.json")):
            project_root, data, bash_state = load_memory(
                task_id, max_tokens=2, encoder=mock_encoder, decoder=mock_decoder
            )
            self.assertEqual(project_root, "")
            self.assertEqual(data, "Decoded text\n(... truncated)")
            self.assertEqual(
                bash_state, bash_state_data
            )  # Verify bash state was loaded

    def test_load_memory_with_tokens(self):
        """Test loading of memory data with token limit"""
        task_id = "test-id"
        memory_data = "# PROJECT ROOT = '/project'\nTest description\n*.py\nfile1.py"
        bash_state_data = {
            "bash_command_mode": {
                "bash_mode": "normal_mode",
                "allowed_commands": "all",
            },
            "file_edit_mode": {"allowed_globs": "all"},
            "write_if_empty_mode": {"allowed_globs": "all"},
            "whitelist_for_overwrite": [],
            "mode": "wcgw",
        }

        mock_encoder = lambda x: self.MockTokens([1, 2, 3])  # Simulate tokenizer output
        mock_decoder = lambda x: "truncated text"  # Return fixed text for any input IDs

        from wcgw.client.memory import get_app_dir_xdg

        app_dir = get_app_dir_xdg()
        memory_dir = os.path.join(app_dir, "memory")
        mock_files = {
            os.path.join(memory_dir, f"{task_id}.txt"): memory_data,
            os.path.join(memory_dir, f"{task_id}_bash_state.json"): json.dumps(
                bash_state_data
            ),
        }

        def mock_open_file(filename, *args, **kwargs):
            content = mock_files.get(filename)
            if content is None:
                raise FileNotFoundError(filename)
            return mock_open(read_data=content)()

        with patch("builtins.open", side_effect=mock_open_file), \
             patch("os.path.exists", lambda x: x.endswith("_bash_state.json")):
            project_root, data, bash_state = load_memory(
                task_id, max_tokens=2, encoder=mock_encoder, decoder=mock_decoder
            )

            # Since encoder returns [1, 2, 3] and decoder returns input unchanged,
            # only the first chunk (after truncation) should be returned plus the truncation message
            self.assertEqual(project_root, "")
            # Our test decoder returns unchanged input, so we get first token plus truncation message
            self.assertEqual(
                data, "truncated text\n(... truncated)"
            )  # Use the actual mock_decoder output
            self.assertEqual(
                bash_state, bash_state_data
            )  # Verify bash state was loaded


if __name__ == "__main__":
    unittest.main()



"""Tests for KnowledgeTransfer functionality in tools.py"""

import os
import tempfile
import unittest
from unittest.mock import patch

from wcgw.client.tools import get_tool_output
from wcgw.types_ import ContextSave


class TestKnowledgeTransfer(unittest.TestCase):
    def setUp(self):
        """Set up test environment before each test case"""
        self.maxDiff = None
        self.test_id = "test_task_1"
        test_dir = os.path.join(os.path.dirname(__file__), "test_files")
        os.makedirs(test_dir, exist_ok=True)
        self.test_file_1 = os.path.join(test_dir, "test1.py")
        self.test_file_2 = os.path.join(test_dir, "test2.py")
        self.test_paths = [self.test_file_1, self.test_file_2]
        self.test_files_content = "file content"
        
        # Setup save_memory mock
        self.memory_patch = patch("wcgw.client.tools.save_memory")
        self.mock_save_memory = self.memory_patch.start()
        self.mock_save_memory.return_value = "/tmp/memory/test_task_1.txt"

    def tearDown(self):
        self.memory_patch.stop()

    def test_knowledge_transfer_basic(self):
        """Test basic knowledge transfer functionality"""
        with patch("wcgw.client.tools.read_files") as mock_read_files:
            mock_read_files.return_value = self.test_files_content

            description = "Test objective. Instructions: Test instructions. Status: Test status. Issues: Test snippets."
            kt_arg = ContextSave(
                id=self.test_id,
                project_root_path="/test",
                description=description,
                relevant_file_globs=self.test_paths
            )

            outputs, cost = get_tool_output(kt_arg, None, 1.0, lambda x, y: ("", 0), None)
            output = outputs[0]

            self.assertTrue(isinstance(output, str))
            self.assertTrue(output.endswith(".txt"))
            
            # Verify save_memory was called correctly
            self.mock_save_memory.assert_called_once()
            call_args = self.mock_save_memory.call_args[0]
            self.assertEqual(call_args[0].id, self.test_id)
            self.assertEqual(call_args[0].project_root_path, "/test") 
            self.assertEqual(call_args[0].description, description)
            self.assertEqual(call_args[0].relevant_file_globs, self.test_paths)

    def test_knowledge_transfer_with_empty_fields(self):
        """Test knowledge transfer with empty fields"""
        with patch("wcgw.client.tools.read_files") as mock_read_files:
            mock_read_files.return_value = self.test_files_content

            kt_arg = ContextSave(
                id=self.test_id,
                project_root_path="/test",
                description="",
                relevant_file_globs=[]
            )

            outputs, cost = get_tool_output(kt_arg, None, 1.0, lambda x, y: ("", 0), None)
            output = outputs[0]

            self.assertTrue(output.endswith(".txt"))
            
            # Verify save_memory was called correctly
            self.mock_save_memory.assert_called_once()
            call_args = self.mock_save_memory.call_args[0]
            self.assertEqual(call_args[0].description, "")
            self.assertEqual(call_args[0].relevant_file_globs, [])

    def test_knowledge_transfer_file_read_failure(self):
        """Test knowledge transfer when file reading fails"""
        with patch("wcgw.client.tools.read_files") as mock_read_files:
            mock_read_files.return_value = "Failed to read files"

            description = "Test fail case"
            kt_arg = ContextSave(
                id=self.test_id,
                project_root_path="/test",
                description=description,
                relevant_file_globs=self.test_paths
            )

            outputs, cost = get_tool_output(kt_arg, None, 1.0, lambda x, y: ("", 0), None)
            output = outputs[0]

            self.assertTrue(output.endswith(".txt"))
            
            # Verify save_memory was called correctly
            call_args = self.mock_save_memory.call_args[0]
            self.assertEqual(call_args[0].id, self.test_id)
            self.assertEqual(call_args[0].description, description)

    def test_knowledge_transfer_with_long_content(self):
        """Test knowledge transfer with long content"""
        with patch("wcgw.client.tools.read_files") as mock_read_files:
            mock_read_files.return_value = self.test_files_content

            long_content = "A" * 10000  # 10KB of content
            description = f"Test with long content: {long_content}"
            kt_arg = ContextSave(
                id=self.test_id,
                project_root_path="/test",
                description=description,
                relevant_file_globs=self.test_paths
            )

            outputs, cost = get_tool_output(kt_arg, None, 1.0, lambda x, y: ("", 0), None)
            output = outputs[0]

            self.assertTrue(output.endswith(".txt"))
            
            # Verify save_memory was called correctly
            call_args = self.mock_save_memory.call_args[0]
            self.assertEqual(call_args[0].id, self.test_id)
            self.assertEqual(call_args[0].description, description)

    def test_knowledge_transfer_with_special_characters(self):
        """Test knowledge transfer with special characters in fields"""
        with patch("wcgw.client.tools.read_files") as mock_read_files:
            mock_read_files.return_value = self.test_files_content

            special_text = 'Test with special chars: \n\t"\'[]{}!'
            description = f"Test with special characters: {special_text}"
            kt_arg = ContextSave(
                id=self.test_id,
                project_root_path="/test",
                description=description,
                relevant_file_globs=self.test_paths
            )

            outputs, cost = get_tool_output(kt_arg, None, 1.0, lambda x, y: ("", 0), None)
            output = outputs[0]

            self.assertTrue(output.endswith(".txt"))
            
            # Verify save_memory was called correctly
            call_args = self.mock_save_memory.call_args[0]
            self.assertEqual(call_args[0].id, self.test_id)
            self.assertEqual(call_args[0].description, description)


if __name__ == "__main__":
    unittest.main()


import unittest
from unittest.mock import patch, MagicMock
from wcgw.client.computer_use import (
    ComputerTool,
    ToolResult,
    ToolError,
    ScalingSource,
    chunks,
)
import time


class TestComputerUse(unittest.TestCase):
    def setUp(self):
        self.computer = ComputerTool()
        self.computer.width = 1920
        self.computer.height = 1080
        self.computer.display_num = 0
        self.computer.docker_image_id = "test_docker_id"
        self.computer._display_prefix = "DISPLAY=:0 "
        self.computer.xdotool = "DISPLAY=:0 xdotool"

    def test_tool_result_operations(self):
        # Test ToolResult creation and combination
        result1 = ToolResult(output="test1", error=None, base64_image=None)
        result2 = ToolResult(output="test2", error="error2", base64_image=None)
        
        # Test addition
        combined = result1 + result2
        self.assertEqual(combined.output, "test1test2")
        self.assertEqual(combined.error, "error2")
        self.assertIsNone(combined.base64_image)

        # Test bool evaluation
        self.assertTrue(bool(result1))
        self.assertFalse(bool(ToolResult()))

    def test_chunks_function(self):
        # Test the chunks utility function
        text = "123456789"
        result = chunks(text, 3)
        self.assertEqual(result, ["123", "456", "789"])
        
        # Test with uneven chunks
        result = chunks(text, 4)
        self.assertEqual(result, ["1234", "5678", "9"])

    @patch("wcgw.client.computer_use.command_run")
    def test_get_screen_info(self, mock_command_run):
        # Mock the shell command output
        mock_command_run.return_value = (0, "1920,1080,0", "")
        
        width, height, display_num = self.computer.get_screen_info()
        
        self.assertEqual(width, 1920)
        self.assertEqual(height, 1080)
        self.assertEqual(display_num, 0)
        self.assertEqual(self.computer._display_prefix, "DISPLAY=:0 ")

    @patch("wcgw.client.computer_use.command_run")
    @patch("os.path.exists")
    @patch("builtins.open")
    def test_mouse_move_action(self, mock_open, mock_exists, mock_command_run):
        mock_command_run.return_value = (0, "", "")
        mock_exists.return_value = True
        mock_open.return_value.__enter__.return_value.read.return_value = b"test_image"
        
        # Test simple mouse move
        result = self.computer(action="mouse_move", coordinate=(100, 100))
        self.assertEqual(result.error, "")
        
        # Test mouse move with click
        result = self.computer(
            action="mouse_move",
            coordinate=(100, 100),
            do_left_click_on_move=True
        )
        self.assertEqual(result.error, "")

    @patch("wcgw.client.computer_use.command_run")
    @patch("os.path.exists")
    @patch("builtins.open")
    def test_typing_actions(self, mock_open, mock_exists, mock_command_run):
        mock_command_run.return_value = (0, "", "")
        mock_exists.return_value = True
        mock_open.return_value.__enter__.return_value.read.return_value = b"test_image"
        
        # Test key action
        result = self.computer(action="key", text="Return")
        self.assertEqual(result.error, "")
        
        # Test type action
        text = "Hello\nWorld"
        result = self.computer(action="type", text=text)
        self.assertEqual(result.error, "")

    def test_invalid_inputs(self):
        # Test invalid action
        with self.assertRaises(ToolError):
            self.computer(action="invalid_action")
        
        # Test missing coordinate for mouse move
        with self.assertRaises(ToolError):
            self.computer(action="mouse_move")
        
        # Test invalid coordinate type
        with self.assertRaises(ToolError):
            self.computer(action="mouse_move", coordinate="invalid")

    def test_scaling_coordinates(self):
        # Test API to computer scaling
        x, y = self.computer.scale_coordinates(ScalingSource.API, 800, 600)
        self.assertTrue(isinstance(x, int))
        self.assertTrue(isinstance(y, int))
        
        # Test computer to API scaling
        x, y = self.computer.scale_coordinates(ScalingSource.COMPUTER, 1920, 1080)
        self.assertTrue(isinstance(x, int))
        self.assertTrue(isinstance(y, int))
        
        # Test out of bounds coordinates
        with self.assertRaises(ToolError):
            self.computer.scale_coordinates(ScalingSource.API, 2000, 2000)

    @patch("wcgw.client.computer_use.command_run")
    @patch("os.path.exists")
    @patch("builtins.open")
    def test_screenshot_functionality(self, mock_open, mock_exists, mock_command_run):
        # Configure mocks
        mock_command_run.return_value = (0, "", "")
        mock_exists.return_value = True
        mock_open.return_value.__enter__.return_value.read.return_value = b"test_image_data"
        mock_open.return_value.__exit__ = MagicMock()
        
        # Test screenshot
        result = self.computer.screenshot()
        self.assertIsNotNone(result.base64_image)
        self.assertEqual(result.error, "")


if __name__ == "__main__":
    unittest.main()



import logging
from contextlib import asynccontextmanager

import anyio
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from starlette.types import Receive, Scope, Send
from starlette.websockets import WebSocket

import mcp_wcgw.types as types

logger = logging.getLogger(__name__)


@asynccontextmanager
async def websocket_server(scope: Scope, receive: Receive, send: Send):
    """
    WebSocket server transport for mcp_wcgw This is an ASGI application, suitable to be
    used with a framework like Starlette and a server like Hypercorn.
    """

    websocket = WebSocket(scope, receive, send)
    await websocket.accept(subprotocol="mcp")

    read_stream: MemoryObjectReceiveStream[types.JSONRPCMessage | Exception]
    read_stream_writer: MemoryObjectSendStream[types.JSONRPCMessage | Exception]

    write_stream: MemoryObjectSendStream[types.JSONRPCMessage]
    write_stream_reader: MemoryObjectReceiveStream[types.JSONRPCMessage]

    read_stream_writer, read_stream = anyio.create_memory_object_stream(0)
    write_stream, write_stream_reader = anyio.create_memory_object_stream(0)

    async def ws_reader():
        try:
            async with read_stream_writer:
                async for message in websocket.iter_json():
                    try:
                        client_message = types.JSONRPCMessage.model_validate(message)
                    except Exception as exc:
                        await read_stream_writer.send(exc)
                        continue

                    await read_stream_writer.send(client_message)
        except anyio.ClosedResourceError:
            await websocket.close()

    async def ws_writer():
        try:
            async with write_stream_reader:
                async for message in write_stream_reader:
                    obj = message.model_dump(
                        by_alias=True, mode="json", exclude_none=True
                    )
                    await websocket.send_json(obj)
        except anyio.ClosedResourceError:
            await websocket.close()

    async with anyio.create_task_group() as tg:
        tg.start_soon(ws_reader)
        tg.start_soon(ws_writer)
        yield (read_stream, write_stream)



import unittest
from unittest.mock import patch, MagicMock
from wcgw.client.common import discard_input, CostData
import sys
import termios
import select


class TestCommon(unittest.TestCase):
    def test_cost_data_model(self):
        """Test CostData model initialization and validation"""
        cost_data = CostData(
            cost_per_1m_input_tokens=0.01,
            cost_per_1m_output_tokens=0.02
        )
        self.assertEqual(cost_data.cost_per_1m_input_tokens, 0.01)
        self.assertEqual(cost_data.cost_per_1m_output_tokens, 0.02)

    @patch('sys.stdin')
    @patch('termios.tcgetattr')
    @patch('termios.tcsetattr')
    @patch('tty.setcbreak')
    @patch('select.select')
    def test_discard_input_with_data(
        self,
        mock_select,
        mock_setcbreak,
        mock_tcsetattr,
        mock_tcgetattr,
        mock_stdin
    ):
        # Mock file descriptor
        mock_stdin.fileno.return_value = 0
        
        # Mock terminal settings
        mock_tcgetattr.return_value = 'old_settings'
        
        # Mock select to indicate there is input first, then no input
        mock_select.side_effect = [
            ([mock_stdin], [], []),  # First call - input available
            ([], [], [])  # Second call - no input
        ]
        
        # Mock reading input
        mock_stdin.read.return_value = 'x'
        
        discard_input()
        
        # Verify terminal settings were properly managed
        mock_tcgetattr.assert_called_once_with(0)
        mock_setcbreak.assert_called_once_with(0)
        mock_tcsetattr.assert_called_once_with(0, termios.TCSADRAIN, 'old_settings')
        
        # Verify input was read
        mock_stdin.read.assert_called_with(1)

    @patch('sys.stdin')
    @patch('termios.tcgetattr')
    @patch('termios.tcsetattr')
    @patch('tty.setcbreak')
    @patch('select.select')
    def test_discard_input_no_data(
        self,
        mock_select,
        mock_setcbreak,
        mock_tcsetattr,
        mock_tcgetattr,
        mock_stdin
    ):
        # Mock file descriptor
        mock_stdin.fileno.return_value = 0
        
        # Mock terminal settings
        mock_tcgetattr.return_value = 'old_settings'
        
        # Mock select to indicate no input
        mock_select.return_value = ([], [], [])
        
        discard_input()
        
        # Verify terminal settings were properly managed
        mock_tcgetattr.assert_called_once_with(0)
        mock_setcbreak.assert_called_once_with(0)
        mock_tcsetattr.assert_called_once_with(0, termios.TCSADRAIN, 'old_settings')
        
        # Verify no input was read
        mock_stdin.read.assert_not_called()

    @patch('sys.stdin')
    @patch('termios.tcgetattr')
    @patch('builtins.print')
    def test_discard_input_error_handling(
        self,
        mock_print,
        mock_tcgetattr,
        mock_stdin
    ):
        # Mock termios error
        mock_tcgetattr.side_effect = termios.error("Mock termios error")
        
        discard_input()
        
        # Verify error was handled gracefully
        mock_print.assert_called_once()
        error_message = mock_print.call_args[0][0]
        self.assertTrue("Warning: Unable to discard input" in error_message)
        self.assertTrue("Mock termios error" in error_message)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch, mock_open, MagicMock
import base64
from wcgw.client.anthropic_client import parse_user_message_special, text_from_editor
from wcgw.client.tools import ImageData
import tempfile
import os
import json
from pathlib import Path


class TestAnthropicClient(unittest.TestCase):
    def setUp(self):
        self.console_mock = MagicMock()
    
    def test_parse_user_message_special_text_only(self):
        msg = "Hello\nThis is a test"
        result = parse_user_message_special(msg)
        self.assertEqual(result["role"], "user")
        self.assertEqual(len(result["content"]), 1)
        self.assertEqual(result["content"][0]["type"], "text")
        self.assertEqual(result["content"][0]["text"], "Hello\nThis is a test")

    @patch("builtins.open")
    @patch("mimetypes.guess_type")
    def test_parse_user_message_special_with_image(self, mock_guess_type, mock_open):
        # Mock image data
        image_data = b"fake_image_data"
        mock_open.return_value.__enter__.return_value.read.return_value = image_data
        mock_guess_type.return_value = ("image/png", None)
        
        msg = "%image test.png\nSome text after"
        result = parse_user_message_special(msg)
        
        # Verify structure
        self.assertEqual(result["role"], "user")
        self.assertEqual(len(result["content"]), 2)
        
        # Verify image block
        self.assertEqual(result["content"][0]["type"], "image")
        self.assertEqual(result["content"][0]["source"]["type"], "base64")
        self.assertEqual(result["content"][0]["source"]["media_type"], "image/png")
        expected_b64 = base64.b64encode(image_data).decode("utf-8")
        self.assertEqual(result["content"][0]["source"]["data"], expected_b64)
        
        # Verify text block
        self.assertEqual(result["content"][1]["type"], "text")
        self.assertEqual(result["content"][1]["text"], "Some text after")

    @patch("builtins.input")
    def test_text_from_editor_direct_input(self, mock_input):
        mock_input.return_value = "Test direct input"
        result = text_from_editor(self.console_mock)
        self.assertEqual(result, "Test direct input")
        
    @patch("tempfile.NamedTemporaryFile")
    @patch("subprocess.run")
    @patch("builtins.input")
    def test_text_from_editor_with_editor(self, mock_input, mock_run, mock_tempfile):
        # Set up mocks
        mock_input.return_value = ""  # Empty input triggers editor
        mock_tempfile.return_value.__enter__.return_value.name = "test.tmp"
        
        # Mock the editor content
        editor_content = "Content from editor"
        with patch("builtins.open", mock_open(read_data=editor_content)):
            result = text_from_editor(self.console_mock)
            
        self.assertEqual(result, editor_content)
        mock_run.assert_called_once()
        
    @patch("os.environ")
    def test_text_from_editor_custom_editor(self, mock_environ):
        # Test with custom editor
        mock_environ.get.return_value = "nano"
        with patch("builtins.input", return_value=""):
            with patch("tempfile.NamedTemporaryFile") as mock_tmp:
                mock_tmp.return_value.__enter__.return_value.name = "test.tmp"
                with patch("subprocess.run") as mock_run:
                    with patch("builtins.open", mock_open(read_data="Written in nano")):
                        result = text_from_editor(self.console_mock)
                        
        self.assertEqual(result, "Written in nano")
        mock_run.assert_called_once_with(["nano", "test.tmp"], check=True)


if __name__ == "__main__":
    unittest.main()



import pytest
from unittest.mock import MagicMock, patch
from wcgw.client.tools import execute_bash, BashCommand, BashInteraction, render_terminal_output

class TestExecuteBash:
    def setup_method(self):
        self.mock_tokenizer = MagicMock()
    
    @patch('wcgw.client.tools.BASH_STATE')
    def test_execute_bash_command(self, mock_bash_state):
        # Setup mock with rendered output
        mock_shell = MagicMock()
        test_output = render_terminal_output("test output\n")
        mock_shell.before = test_output[0] if test_output else ""
        mock_shell.linesep = "\n"
        
        # Setup BASH_STATE
        mock_bash_state.shell = mock_shell
        mock_bash_state.state = "repl"
        mock_bash_state.pending_output = ""
        mock_bash_state.update_cwd.return_value = "/test/path"
        
        # Create command and execute
        command = BashCommand(command="echo test")
        output, cost = execute_bash(self.mock_tokenizer, command, max_tokens=100, timeout_s=1)
        
        assert "test output" in output.strip()
    
    @patch('wcgw.client.tools.BASH_STATE')
    def test_execute_bash_interaction(self, mock_bash_state):
        mock_shell = MagicMock()
        test_output = render_terminal_output("test output\n")
        mock_shell.before = test_output[0] if test_output else ""
        mock_shell.linesep = "\n"
        
        mock_bash_state.shell = mock_shell
        mock_bash_state.state = "repl"
        mock_bash_state.pending_output = ""
        mock_bash_state.update_cwd.return_value = "/test/path"
        
        interaction = BashInteraction(send_text="test input")
        output, cost = execute_bash(self.mock_tokenizer, interaction, max_tokens=100, timeout_s=1)
        
        assert "test output" in output.strip()
    
    @patch('wcgw.client.tools.BASH_STATE')
    def test_execute_bash_pending(self, mock_bash_state):
        """Test execution when a command is already running"""
        mock_bash_state.state = "pending"
        
        command = BashCommand(command="echo test")
        with pytest.raises(ValueError, match="A command is already running"):
            execute_bash(self.mock_tokenizer, command, max_tokens=100, timeout_s=1)
    
    @patch('wcgw.client.tools.BASH_STATE')
    def test_execute_bash_large_output(self, mock_bash_state):
        mock_shell = MagicMock()
        large_output = "large output\n" * 10
        test_output = render_terminal_output(large_output)
        mock_shell.before = test_output[0] if test_output else ""
        mock_shell.linesep = "\n"
        
        mock_bash_state.shell = mock_shell
        mock_bash_state.state = "repl"
        mock_bash_state.pending_output = ""
        mock_bash_state.update_cwd.return_value = "/test/path"
        
        command = BashCommand(command="long_running_command")
        output, cost = execute_bash(self.mock_tokenizer, command, max_tokens=100, timeout_s=1)
        
        assert "large output" in output.strip()


"""
This module provides simpler types to use with the server for managing prompts
and tools.
"""

from pydantic import BaseModel

from mcp_wcgw.types import (
    ServerCapabilities,
)


class InitializationOptions(BaseModel):
    server_name: str
    server_version: str
    capabilities: ServerCapabilities



import os
import sys
from contextlib import asynccontextmanager

import anyio
import anyio.lowlevel
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from anyio.streams.text import TextReceiveStream
from pydantic import BaseModel, Field

import mcp_wcgw.types as types

# Environment variables to inherit by default
DEFAULT_INHERITED_ENV_VARS = (
    [
        "APPDATA",
        "HOMEDRIVE",
        "HOMEPATH",
        "LOCALAPPDATA",
        "PATH",
        "PROCESSOR_ARCHITECTURE",
        "SYSTEMDRIVE",
        "SYSTEMROOT",
        "TEMP",
        "USERNAME",
        "USERPROFILE",
    ]
    if sys.platform == "win32"
    else ["HOME", "LOGNAME", "PATH", "SHELL", "TERM", "USER"]
)


def get_default_environment() -> dict[str, str]:
    """
    Returns a default environment object including only environment variables deemed
    safe to inherit.
    """
    env: dict[str, str] = {}

    for key in DEFAULT_INHERITED_ENV_VARS:
        value = os.environ.get(key)
        if value is None:
            continue

        if value.startswith("()"):
            # Skip functions, which are a security risk
            continue

        env[key] = value

    return env


class StdioServerParameters(BaseModel):
    command: str
    """The executable to run to start the server."""

    args: list[str] = Field(default_factory=list)
    """Command line arguments to pass to the executable."""

    env: dict[str, str] | None = None
    """
    The environment to use when spawning the process.

    If not specified, the result of get_default_environment() will be used.
    """


@asynccontextmanager
async def stdio_client(server: StdioServerParameters):
    """
    Client transport for stdio: this will connect to a server by spawning a
    process and communicating with it over stdin/stdout.
    """
    read_stream: MemoryObjectReceiveStream[types.JSONRPCMessage | Exception]
    read_stream_writer: MemoryObjectSendStream[types.JSONRPCMessage | Exception]

    write_stream: MemoryObjectSendStream[types.JSONRPCMessage]
    write_stream_reader: MemoryObjectReceiveStream[types.JSONRPCMessage]

    read_stream_writer, read_stream = anyio.create_memory_object_stream(0)
    write_stream, write_stream_reader = anyio.create_memory_object_stream(0)

    process = await anyio.open_process(
        [server.command, *server.args],
        env=server.env if server.env is not None else get_default_environment(),
        stderr=sys.stderr,
    )

    async def stdout_reader():
        assert process.stdout, "Opened process is missing stdout"

        try:
            async with read_stream_writer:
                buffer = ""
                async for chunk in TextReceiveStream(process.stdout):
                    lines = (buffer + chunk).split("\n")
                    buffer = lines.pop()

                    for line in lines:
                        try:
                            message = types.JSONRPCMessage.model_validate_json(line)
                        except Exception as exc:
                            await read_stream_writer.send(exc)
                            continue

                        await read_stream_writer.send(message)
        except anyio.ClosedResourceError:
            await anyio.lowlevel.checkpoint()

    async def stdin_writer():
        assert process.stdin, "Opened process is missing stdin"

        try:
            async with write_stream_reader:
                async for message in write_stream_reader:
                    json = message.model_dump_json(by_alias=True, exclude_none=True)
                    await process.stdin.send((json + "\n").encode())
        except anyio.ClosedResourceError:
            await anyio.lowlevel.checkpoint()

    async with (
        anyio.create_task_group() as tg,
        process,
    ):
        tg.start_soon(stdout_reader)
        tg.start_soon(stdin_writer)
        yield read_stream, write_stream



import unittest
import tempfile
from pathlib import Path
import os
import shutil
from wcgw.client.repo_ops.display_tree import DirectoryTree

class TestDirectoryTree(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory for testing
        self.test_dir = tempfile.mkdtemp()
        self.root_path = Path(self.test_dir)

        # Create a sample directory structure
        # root/
        #   ├── dir1/
        #   │   ├── file1.txt
        #   │   └── file2.txt
        #   ├── dir2/
        #   │   └── subdir/
        #   │       └── file3.txt
        #   └── file4.txt

        # Create directories
        dir1 = self.root_path / "dir1"
        dir1.mkdir()
        dir2 = self.root_path / "dir2"
        dir2.mkdir()
        subdir = dir2 / "subdir"
        subdir.mkdir()

        # Create files
        (dir1 / "file1.txt").write_text("content1")
        (dir1 / "file2.txt").write_text("content2")
        (subdir / "file3.txt").write_text("content3")
        (self.root_path / "file4.txt").write_text("content4")

        self.tree = DirectoryTree(self.root_path)

    def tearDown(self):
        # Clean up the temporary directory
        shutil.rmtree(self.test_dir)

    def test_initialization(self):
        # Test initialization with valid directory
        self.assertEqual(self.tree.root, self.root_path)
        self.assertEqual(self.tree.max_files, 10)
        self.assertEqual(len(self.tree.expanded_files), 0)
        self.assertEqual(len(self.tree.expanded_dirs), 0)

        # Test initialization with non-existent directory
        with self.assertRaises(ValueError):
            DirectoryTree(Path("/nonexistent/path"))

        # Test initialization with file instead of directory
        file_path = self.root_path / "file4.txt"
        with self.assertRaises(ValueError):
            DirectoryTree(file_path)

    def test_expand(self):
        # Test expanding a valid file
        self.tree.expand("file4.txt")
        self.assertIn(self.root_path / "file4.txt", self.tree.expanded_files)
        self.assertIn(self.root_path, self.tree.expanded_dirs)

        # Test expanding a nested file
        self.tree.expand("dir1/file1.txt")
        self.assertIn(self.root_path / "dir1" / "file1.txt", self.tree.expanded_files)
        self.assertIn(self.root_path / "dir1", self.tree.expanded_dirs)

        # Test expanding a deeply nested file
        self.tree.expand("dir2/subdir/file3.txt")
        self.assertIn(self.root_path / "dir2" / "subdir" / "file3.txt", self.tree.expanded_files)
        self.assertIn(self.root_path / "dir2", self.tree.expanded_dirs)
        self.assertIn(self.root_path / "dir2" / "subdir", self.tree.expanded_dirs)

        # Test expanding non-existent file
        with self.assertRaises(ValueError):
            self.tree.expand("nonexistent.txt")

        # Test expanding a directory (should fail)
        with self.assertRaises(ValueError):
            self.tree.expand("dir1")

        # Test path traversal attempt
        with self.assertRaises(ValueError):
            self.tree.expand("../outside.txt")

    def test_list_directory(self):
        # Test listing root directory
        contents = self.tree._list_directory(self.root_path)
        self.assertEqual(len(contents), 3)  # dir1, dir2, file4.txt
        self.assertTrue(all(isinstance(p, Path) for p in contents))

        # Verify directories come before files
        self.assertTrue(contents[0].is_dir())
        self.assertTrue(contents[1].is_dir())
        self.assertTrue(contents[2].is_file())

        # Test listing subdirectory
        dir1_contents = self.tree._list_directory(self.root_path / "dir1")
        self.assertEqual(len(dir1_contents), 2)  # file1.txt, file2.txt
        self.assertTrue(all(p.is_file() for p in dir1_contents))

    def test_count_hidden_items(self):
        # Test with no expanded items
        hidden_files, hidden_dirs = self.tree._count_hidden_items(self.root_path, [])
        self.assertEqual(hidden_files, 1)  # file4.txt
        self.assertEqual(hidden_dirs, 2)   # dir1, dir2

        # Test with some items shown
        shown_items = [
            self.root_path / "dir1",
            self.root_path / "file4.txt"
        ]
        hidden_files, hidden_dirs = self.tree._count_hidden_items(self.root_path, shown_items)
        self.assertEqual(hidden_files, 0)  # file4.txt is shown
        self.assertEqual(hidden_dirs, 1)   # only dir2 is hidden

    def test_display(self):
        # Test initial display (no expansions)
        initial_display = self.tree.display()
        self.assertIn(str(self.root_path), initial_display)
        self.assertIn("2 directories and 1 file hidden", initial_display)

        # Test display with expanded file
        self.tree.expand("file4.txt")
        expanded_display = self.tree.display()
        self.assertIn("file4.txt", expanded_display)
        self.assertIn("2 directories hidden", expanded_display)

        # Test display with expanded nested file
        self.tree.expand("dir1/file1.txt")
        nested_display = self.tree.display()
        self.assertIn("dir1", nested_display)
        self.assertIn("file1.txt", nested_display)
        self.assertIn("1 file hidden", nested_display)  # file2.txt is hidden

        # Test display with all files in dir1 expanded
        self.tree.expand("dir1/file2.txt")
        full_display = self.tree.display()
        self.assertIn("file2.txt", full_display)
        self.assertNotIn("files hidden", full_display)  # all files in dir1 are shown

        # Test display with deeply nested file
        self.tree.expand("dir2/subdir/file3.txt")
        deep_display = self.tree.display()
        self.assertIn("subdir", deep_display)
        self.assertIn("file3.txt", deep_display)
        
if __name__ == "__main__":
    unittest.main()